"""
This module get all the functions related to the transformation of the recordset on the Javascript part.
Each of those functions are in charge of transforming a recordSet in the front end (which means that this cannot be
seen on the Python side in the scripting interface). The best way to test those changes is to use a report and set the DEBUG flag.
"""


class JsRecFunc(object):
  """
  This class cannot be used directly to format a record as the mandatory parameters are not defined and are set to None.
  Anyway as an interface this will give you the different information which have to be defined to create a new js function.
  The objective in this logic is to centralise all the js functions used in the front end in this folder in order
  to limit the use and standardise the implementation.

  ## The main parameters ##

  - alias, The alias name used in the Python layer to point to this function
  - params, The parameters name available in the Javascript function (data is always passed)
  - value, The content of the function.

  ## Javascript function ##

  A javascript function, represented as a class in the framework, is always trying to transform a data variable
  which is defined as a recordSet to a result variable which should be also a recordSet. The purpose of having this
  will ensure that functions can:

    1. Transform the recordSet in a specific manner
    2. Be shared in the different components
    3. Be put together sum(count(...))

  """
  alias = None
  params = None
  value = None

  @staticmethod
  def extendArgs(category, originParams, newCols): return originParams

  @staticmethod
  def extendColumns(jsSchema, params):
    raise Exception("Method extendColumns should be overriden")


class JsRowBuckets(JsRecFunc):
  """
  """
  @staticmethod
  def extendArgs(category, originParams, newCols):
    originParams[1] += newCols
    return originParams

  @staticmethod
  def extendColumns(jsSchema, params):
    pass

  alias = "row-buckets"
  params = ("allGroups", "seriesNames")
  value = '''
    var groupRowsIds = {}; var groupRows = {};
    data.forEach(function(rec, i){
      var inBuckets = {}; for(var g in allGroups){inBuckets[g] = null};
      for(var g in allGroups){
        for(var col in allGroups[g]){if(allGroups[g][col].indexOf(rec[col]) >= 0){
          inBuckets[g] = true} else {inBuckets[g] = false; break}}}
      for(var g in inBuckets){
        if (inBuckets[g]) { 
            if (g in groupRowsIds) {groupRowsIds[g].push(i);groupRows[g].push(rec)} 
            else {groupRowsIds[g] = [i]; groupRows[g] = [rec]}}
      }; result.push(rec)});
      
    for(var g in groupRows){
      var row = {'_system': true}; var text = g;
      for(var col in allGroups[g]){row[col] = text; text=''}; seriesNames.forEach(function(v){row[v] = 0}); 
      groupRows[g].forEach(function(rec){ seriesNames.forEach(function(v){row[v] += rec[v]})});
      result.push(row)}'''


class JsRowTotal(JsRecFunc):
  @staticmethod
  def extendArgs(category, originParams, newCols):
    originParams[0] += newCols
    return originParams

  @staticmethod
  def extendColumns(jsSchema, params):
    pass

  alias = "row-total"
  params = ("seriesNames", "rowDefinition")
  value = '''
    seriesNames.forEach(function(v){rowDefinition[v] = 0});
    data.forEach(function(rec){
      if(!rec['_system']){seriesNames.forEach(function(v){rowDefinition[v] += rec[v]})};
      result.push(rec);
    }); result.push(rowDefinition);
    '''


class JsAll(JsRecFunc):
  @staticmethod
  def extendColumns(jsSchema, params):
    if params[0] is not None and params[1] is not None:
      jsSchema['keys'] |= set(params[0])
      jsSchema['values'] |= set(params[1])

  alias = "all"
  params = ("keys", "values")
  value = 'result = data'


class JsSum(JsRecFunc):
  """
  """

  @staticmethod
  def extendColumns(jsSchema, params):
    if params[0] is not None and params[1] is not None:
      jsSchema['keys'] |= set(params[0])
      jsSchema['values'] |= set(params[1])

  alias = "sum"
  params = ("keys", "values", "xOrder")
  value = '''
    if ((keys == null) || (values == null)){result = data}
    else{
      var temp = {}; var order = []; 
      if (Array.isArray(data)){
        data.forEach( function(rec){ 
          var aggKey = []; keys.forEach(function(k){ aggKey.push(rec[k])}); 
          var newKey = aggKey.join("#"); if (!(newKey in temp)) {order.push(newKey)};
          if (!(newKey in temp)) {temp[newKey] = {}};
          values.forEach(function(v) {if (!(v in temp[newKey])) {temp[newKey][v] = rec[v]} else {temp[newKey][v] += rec[v]}})})}; 
      if(Array.isArray(xOrder)){order = xOrder};
      order.forEach(function(label) {
        var rec = {}; var splitKey = label.split("#");
        keys.forEach(function(k, i) {rec[k] = splitKey[i];});
        for(var v in temp[label]) {rec[v] = temp[label][v]};
        result.push(rec)})}'''


class JsPercentage(JsRecFunc):
  """
  """

  @staticmethod
  def extendColumns(jsSchema, params):
    if params[0] is not None and params[1] is not None:
      jsSchema['keys'] |= set(params[0])
      jsSchema['values'] |= set(params[1])

  alias = "percentage"
  params = ("keys", "values")
  value = ''' 
    if ((keys == null) || (values == null)){result = data}
    else{
      var temp = {}; var order = []; var sumPerSeries = {};
      data.forEach( function(rec) { 
        var aggKey = []; keys.forEach(function(k){ aggKey.push(rec[k])}); 
        var newKey = aggKey.join("#"); if (!(newKey in temp)) {order.push(newKey)};
        if (!(newKey in temp)) {temp[newKey] = {}};
        values.forEach(function(v) {
          if (!(v in sumPerSeries)) {sumPerSeries[v] = rec[v]} else {sumPerSeries[v] += rec[v]};
          if (!(v in temp[newKey])) {temp[newKey][v] = rec[v]} 
          else {temp[newKey][v] += rec[v]}})});
      order.forEach(function(label) {
        var rec = {}; var splitKey = label.split("#");
        keys.forEach(function(k, i) {rec[k] = splitKey[i]});
        for(var v in temp[label]) {rec[v] = 100 * (temp[label][v] / sumPerSeries[v])};
        result.push(rec)})}'''


class JsOperations(JsRecFunc):
  """
  This function will aggregate the different values for each series according to a shcema defined in a Python
  dictionary in the last position of the tuple.

  Example
  aggFnc=('aggregation', ['direction'], values, {'dn': 'sum', 'Date': 'count'}),

  :return:
  """

  @staticmethod
  def extendArgs(category, originParams, newCols):
    """
    This function will update the function argument according to the mode defined by the user. Indeed some properties can be received to validate the accuracy of the data.
    Those data should be added to the different transformation functions and the columns should be passed to the final object.
    This function will ensure that by activating the mode the columns will be automatically added to the aggregated data.

    :return: The update set of columns to be considered in the Javascript function
    """
    if category == 'age':
      originParams[1] = originParams[1] + newCols
      for c in newCols:
        originParams[2][c] = 'sum'
    return originParams

  @staticmethod
  def extendColumns(jsSchema, params):
    """

    """
    if params[0] is not None and params[1] is not None:
      jsSchema['keys'] |= set(params[0])
      jsSchema['values'] |= set(params[1])

  alias = "aggregation"
  params = ("keys", "values", "operations")
  value = '''
    var temp = {};
    var order = [];
    data.forEach( function(rec) { 
      var aggKey = []; keys.forEach(function(k){ aggKey.push( rec[k])}); var newKey = aggKey.join("#"); order.push(newKey);
      if (!(newKey in temp)) {temp[newKey] = {}};
      values.forEach(function(v) {
        if (operations[v] === undefined){ if (!(v in temp[newKey])) {temp[newKey][v] = 1} else {temp[newKey][v] += 1} }
        else if (operations[v] == 'sum') {if (!(v in temp[newKey])) {temp[newKey][v] = rec[v]} else {temp[newKey][v] += rec[v]}}
        else if (operations[v] == 'count') {if (!(v in temp[newKey])) {temp[newKey][v] = 1} else {temp[newKey][v] += 1}}
      })}); 
    order.forEach(function(label) {
      var rec = {}; var splitKey = label.split("#");
      keys.forEach(function(k, i) {rec[k] = splitKey[i];});
      for(var v in temp[label]) {rec[v] = temp[label][v]};
      result.push(rec)})'''


class JsCount(JsRecFunc):
  params = ("keys", "values")
  value = '''
    var temp = {}; var order = [];
    data.forEach(function(rec){ 
      var aggKey = []; keys.forEach(function(k){aggKey.push(rec[k])}); var newKey = aggKey.join("#"); 
      if(!(newKey in temp)){order.push(newKey); temp[newKey] = {}};
      values.forEach(function(v){if (!(v in temp[newKey])){
        temp[newKey][v] = rec[v]; temp[newKey][v +"_count"] = 1} else{temp[newKey][v +"_count"] += 1}})}); 
    order.forEach(function(label){
      var rec = {}; var splitKey = label.split("#");
      keys.forEach(function(k, i){rec[k] = splitKey[i]});
      for(var v in temp[label]){rec[v] = temp[label][v]};
      result.push(rec)})'''


class JsCountSum(JsRecFunc):
  params = ("keys", "values")
  value = '''
    var temp = {}; var order = [];
    data.forEach(function(rec){ 
      var aggKey = []; keys.forEach(function(k){aggKey.push(rec[k])}); var newKey = aggKey.join("#"); 
      if(!(newKey in temp)){order.push(newKey); temp[newKey] = {}};
      values.forEach(function(v){
        if (!(v in temp[newKey])){
          temp[newKey][v] = rec[v]; temp[newKey][v +"_count"] = 1; temp[newKey][v +"_min"] = rec[v]; 
          temp[newKey][v +"_max"] = rec[v]} 
        else{
          if(rec[v] > temp[newKey][v +"_max"]){temp[newKey][v +"_max"] = rec[v]};
          if(rec[v] < temp[newKey][v +"_min"]){temp[newKey][v +"_min"] = rec[v]};
          temp[newKey][v] += rec[v];temp[newKey][v +"_count"] += 1}})}); 
    order.forEach(function(label){
      var rec = {}; var splitKey = label.split("#");
      keys.forEach(function(k, i){rec[k] = splitKey[i]});
      for(var v in temp[label]){rec[v] = temp[label][v]};
      result.push(rec)})'''


class JsTop(object):
  """
  """

  @staticmethod
  def extendColumns(jsSchema, params): pass

  alias = "top"
  params = ("countItems", "value", "sortType")
  value = '''
    var tmpRec = {};
    data.forEach(function(rec){
      if(tmpRec[rec[value]] === undefined){ tmpRec[rec[value]] = [rec] } else {tmpRec[rec[value]].push(rec)}});
    
    var result = []; 
    Object.keys(tmpRec).sort().forEach(function(key){
      tmpRec[key].forEach(function(rec){result.push(rec)})});
    
    if (sortType == 'desc'){ result = result.slice(-countItems)}
    else {result = result.slice(0, countItems)}
    '''


class JsCountDistinct(object):
  """
  Return the distinct counts of element in a list of columns. This function will return a list of dictionaries
  with the following structure {'column': '', 'count_distinct': 0}

  :return: A new recordSet with the properties of the requested keys
  """
  params = ("keys", )
  value = '''
    var temp = {}; keys.forEach(function(k){temp[k] = {}});
    data.forEach(function(rec){keys.forEach(function(k){temp[k][rec[k]] = 1})}); 
    for(var col in temp){
      var dCount = Object.keys(temp[col]).length; 
      result.push({'column': col, 'count': dCount, 'distinct': true, 'values': Object.keys(temp[col])})}'''


class JsCountAll(object):
  """
  Function to produce KPI on an original recordSet. This function will create a new recordSet based on the selected
  columns of the original data source.

  :return: A new recordSet with the properties of the requested keys
  """
  params = ("keys", )
  value = '''
    var temp = {}; var order= [];
    data.forEach(function(rec){ 
      keys.forEach(function(k){
        var aggKey = k +"#"+ rec[k]; if(!(aggKey in temp)){order.push(aggKey); temp[aggKey] = 1} else{temp[aggKey] += 1}})}); 
    order.forEach(function(label){
      var keys = label.split("#"); var rec = {'column': keys[0], 'value': keys[1], 'count': temp[label], 'distinct': false};
      result.push(rec)})'''


class JsRename(object):
  """
  Function to remap some columns in the recordSet. The renaming is done based on the input parameter.
  The parameter passed in this function is a dictionary with as keys the existing column names and value the new column.

  :return: The Js recordSet with the new columns in each record. The original keys will be removed
  """
  alias = "rename"
  params = ("colsWithName", )
  value = '''
    data.forEach(function(rec){ 
      for(var col in colsWithName){rec[colsWithName[col]] = rec[col]; delete rec[col]; result.push(rec)}})'''


class JsExtend(object):
  """
  Function to add some predefined entries to each records in the RecordSet
  The parameter passed in the function call should be a dictionary with as keys the columns to be added to the original record
  and the value {'static': {}, 'dynamic': {}}
  In case of key clashes the values will be replaced.

  :return: A new Js recordSet with the extra columns
  """
  alias = "extend"
  params = ('values', 'recKey')
  value = '''
    if (Array.isArray(data)){
      if(recKey == undefined){
        data.forEach(function(rec, i){ 
          var newRec = Object.assign(rec, values.static);
          if (i in values.dynamic) {newRec = Object.assign(newRec, values.dynamic[i])};
          result.push(newRec)})}
      else{
        data.forEach(function(rec, i){ 
          var newRec = Object.assign(rec, values.static);
          if (i in values.dynamic) {newRec = Object.assign(newRec, values.dynamic[i])};
          result.push(newRec);
          var subValues = rec[recKey]; result[recKey] = [];
          subValues.forEach(function(row, j){ 
            var newRec = Object.assign(row, values.static);
            if (i in values.dynamic) {newRec = Object.assign(newRec, values.dynamic[i])};
            result[recKey].push(newRec)})  
        })
      }} 
    else {result = data}
  '''


class JsExtendDataSet(object):
  """
  :return: A new Js recordSet with the extra columns in the datasets section
  """
  alias = "extend-dataset"
  params = ("values", 'recKey')
  value = '''
    var records; var recResults; 
    if(recKey == undefined){records = data; recResults = result} 
    else {records = data[recKey];result[recKey] = [];recResults = result[recKey];
      for(var k in data) {if (k != recKey) {result[k] = data[k]}}};
    records.forEach(function(rec, i) { 
      var newRec = Object.assign(rec, values.static);
      if (i in values.dynamic) {newRec = Object.assign(newRec, values.dynamic[i])};
      recResults.push(newRec)})'''


class JsFilter(object):
  """
  Filter the different records in a recordSet from the definition given as a parameter.
  The filters definition is based on a dictionary as keys the column names. Each records should have the given columns.
  All records which do not match the rules will not be considered

  :return: A new JS dictionary with only the selected lines
  """
  pmts = ("filterCols", )
  content = ''' 
    filters = {};
    filterCols.forEach(function(rec){  
      if (filters[rec['colName']] === undefined){
        filters[rec['colName']] = {val: [], op: rec['op'], allIfEmpty: rec['allIfEmpty']}};
      if(Array.isArray(rec['val'])){ filters[rec['colName']].val = filters[rec['colName']].val.concat(rec['val'])}
      else if (filters[rec['colName']].val.indexOf(rec['val']) < 0) { filters[rec['colName']].val.push(rec['val'])}});
    data.forEach( function(rec) {  
      var isValid = true; 
      for (var col in filters) {
        var opType = filters[col]['op']; 
        if (opType == '='){if(filters[col].val.indexOf(rec[col]) < 0){if ( (filters[col].val != '') && filters[col].allIfEmpty ) {isValid = false; break}}}
        else if (opType == 'in'){ 
          if(filters[col].val.indexOf(rec[col]) < 0){isValid = false; break}
          else if ((filters[col].val.length == 0) && !filters[col].allIfEmpty) {isValid = false; break}}
        else if (opType == 'above'){if(filters[col].val[0] > rec[col]){isValid = false; break}}
        else if (opType == 'below'){if(filters[col].val[0] < rec[col]){isValid = false; break}}
        else if (opType == 'between'){if((filters[col].val[0] > rec[col]) || (filters[col].val[1] < rec[col])){isValid = false; break}}
    }; if (isValid) {result.push(rec)}})'''


class JsIntensity(object):
  """

  """
  alias = "intensity"
  params = ("cols",)
  value = '''
    stats = {};
    cols.forEach(function(col){stats[col] = {min: null, max: null}});
    data.forEach(function(rec){
      cols.forEach(function(col){
        if((stats[col].max == null) || (rec[col] > stats[col].max) ) { stats[col].max = rec[col]};
        if((stats[col].min == null) || (rec[col] < stats[col].min) ) { stats[col].min = rec[col]}})});
    data.forEach(function(rec){
      cols.forEach(function(col){rec[col + ".intensity.min"] = stats[col].min; rec[col + ".intensity.max"] = stats[col].max});
      result.push(rec)});
    '''

  @staticmethod
  def extendColumns(jsSchema, params):
    pass


class JsToUrl(object):
  alias = "dictToUrl"
  value = '''
    var tmpResults = [];
    for(var k in data["pmts"]){tmpResults.push(k +"="+ data["pmts"][k])}; 
    result = tmpResults.join("&");
    if (typeof data["anchor"] !== 'undefined'){result = result +"#"+ data["anchor"]}
    '''




from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsObject


class JsWhile(object):

  def __init__(self, jsPivot, options=None):

    self.options = {"var": 'i'}
    if options is not None:
      self.options.update(options)
    self.__jsFncs, self.__next = [], None
    self.__pivot = jsPivot

  def next(self, rule):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param rule:
    """
    self.__next = rule
    return self

  def fncs(self, jsFncs, reset=True):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFncs:
    :param reset:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    if reset:
      self.__jsFncs = jsFncs
    else:
      self.__jsFncs.extend(jsFncs)
    return self
#
  def toStr(self):
    print(self.__pivot)
    if self.__next is None:
      raise Exception("next() function must be used to avoid infinite loops !!")

    fncs = JsUtils.jsConvertFncs(self.__jsFncs, toStr=True)
    return "while(%s){%s; %s}" % (self.__pivot, fncs, self.__next)


class JsWhileIterable(object):

  def __init__(self, jsIterable, options=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsIterable:
    :param options: Dictionary. While options
    """
    self.__js_it = jsIterable
    self.options = {"var": 'x'}
    if options is not None:
      self.options.update(options)
    self.options['it'] = JsUtils.jsConvertData(self.__js_it, None)

  @property
  def var(self):
    """
    Description:
    -----------
`   Return the variable reference for this loop
    """
    return self.options['var']

  @var.setter
  def var(self, value):
    """
    Description:
    -----------
`   Return the variable reference for this loop

    Attributes:
    ----------
    :param value: String. The value reference for the JavaScript variable
    """
    self.options['var'] = value

  @property
  def value(self):
    """
    Description:
    -----------
    return the value during the while loop
    """
    return JsObject.JsObject.get("%(it)s[%(var)s]" % self.options)

  def fncs(self, jsFncs, reset=True):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFncs: Array. The PyJs functions
    :param reset: Boolean. Reset the JavaScript functions for this loop
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    if reset:
      self.__jsFncs = jsFncs
    else:
      self.__jsFncs.extend(jsFncs)
    return self

  def toStr(self):
    """

    """
    self.options['jsFncs'] = JsUtils.jsConvertFncs(self.__jsFncs, toStr=True)
    return "var %(var)s = 0; while(%(it)s[%(var)s]){%(jsFncs)s; %(var)s++}" % self.options




from epyk.core.html import Html
from epyk.core.js.packages import packageImport
from epyk.core.js.packages import JsDatatable
from epyk.core.js import JsUtils

from epyk.core.data.DataClass import DataClass
from epyk.core.data.DataClass import DataEnum

# The list of CSS classes
from epyk.core.css.styles import GrpClsTable


# External Datatable extensions added on demand to add some extra features
# Details of the different extensions are available on the different websites
# https://datatables.net/extensions/
extensions = {
  'rowsGroup': {'jsImports': ['datatables-rows-group']},
}


class Table(Html.Html):
  requirements = ('datatables', )
  name = 'Table'

  def __init__(self, report, records, width, height, htmlCode, options, profile):
    data, columns, self.__config = [], [], None
    super(Table, self).__init__(report, [], htmlCode=htmlCode, css_attrs={"width": width, "height": height}, profile=profile)
    if records is not None:
      self.config.data = records

  @property
  def style(self):
    if self._styleObj is None:
      self._styleObj = GrpClsTable.Datatable(self)
    return self._styleObj

  @property
  def tableId(self):
    """
    Return the Javascript variable of the chart
    """
    return "%s_obj" % self.htmlCode

  @property
  def config(self):
    if self.__config is None:
      self.__config = TableConfig(self._report)
    return self.__config

  def get_column(self, by_title):
    for c in self.config._attrs.get('columns', []):
      if c.title == by_title:
        return c

    return None

  @property
  def js(self):
    """
    Return the Javascript internal object

    :return: A Javascript object

    :rtype: JsDatatable.DatatableAPI
    """
    if self._js is None:
      self._js = JsDatatable.DatatableAPI(self._report, selector=self.tableId, setVar=False, parent=self)
    return self._js

  def build(self, data=None, options=None, profile=False):
    if data:
      return JsUtils.jsConvertFncs([self.js.clear(), self.js.rows.add(data, update=True)], toStr=True) # self.js.rows.add(data)

    return 'var %s = %s.DataTable(%s)' % (self.tableId, self.dom.jquery.varId, self.config)

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return "<table %s></table>" % (self.get_attrs(pyClassNames=self.style.get_classes()))


class EnumStyleOptions(DataEnum):
  js_conversion = True

  def __wrap(self, name,  header_only=False, body_only=False):
    """

    https://datatables.net/manual/styling/classes
    """
    if header_only:
      return self.set('dt-head-%s' % name)

    if body_only:
      return self.set('dt-body-%s' % name)

    return self.set('dt-%s' % name)

  def left(self, header_only=False, body_only=False):
    """

    https://datatables.net/manual/styling/classes
    """
    return self.__wrap('left', header_only, body_only)

  def right(self, header_only=False, body_only=False):
    """

    https://datatables.net/manual/styling/classes
    """
    return self.__wrap('right', header_only, body_only)

  def center(self, header_only=False, body_only=False):
    """

    https://datatables.net/manual/styling/classes
    """
    return self.__wrap('center', header_only, body_only)

  def justify(self, header_only=False, body_only=False):
    """

    https://datatables.net/manual/styling/classes
    """
    return self.__wrap('justify', header_only, body_only)

  def nowrap(self, header_only=False, body_only=False):
    """

    https://datatables.net/manual/styling/classes
    """
    return self.__wrap('nowrap', header_only, body_only)


class ColumnDef(DataClass):

  @property
  def targets(self):
    """
    Description:
    -----------
    The columnDefs option allows a column definition object to be defined and then assigned to one or more columns in a DataTable, regardless of the order of the column definitions array, or the order of the columns in the table.

    Related Pages:

			https://datatables.net/reference/option/columnDefs.targets
    """
    return self._attrs["targets"]

  @targets.setter
  def targets(self, val):
    self._attrs["targets"] = val

  @property
  def visible(self):
    """
    Description:
    -----------
    Enable or disable the display of this column.

    Related Pages:

			https://datatables.net/reference/option/columns.visible
    """
    return self._attrs["visible"]

  @visible.setter
  def visible(self, val):
    self._attrs["visible"] = val

  @property
  def searchable(self):
    """
    Description:
    -----------
    Enable or disable search on the data in this column.

    Related Pages:

			https://datatables.net/reference/option/columns.searchable
    """
    return self._attrs["searchable"]

  @searchable.setter
  def searchable(self, val):
    self._attrs["searchable"] = val

  @property
  def orderData(self):
    """
    Description:
    -----------
    Define multiple column ordering as the default order for a column.

    Related Pages:

			https://datatables.net/reference/option/columns.orderData
    """
    return self._attrs["orderData"]

  @orderData.setter
  def orderData(self, val):
    self._attrs["orderData"] = val


class Column(DataClass):

  @property
  def cellType(self):
    """
    Description:
    -----------
    Change the cell type created for the column - either TD cells or TH cells.

    Related Pages:

			https://datatables.net/reference/option/columns.cellType
    """
    return self._attrs["cellType"]

  @cellType.setter
  def cellType(self, val):
    self._attrs["cellType"] = val

  @property
  def className(self):
    """
    Description:
    -----------
    Quite simply this option adds a class to each cell in a column, regardless of if the table source is from DOM, Javascript or Ajax. This can be useful for styling columns.

    Related Pages:

			https://datatables.net/reference/option/columns.className
    https://datatables.net/manual/styling/classes

    :rtype: EnumStyleOptions
    """
    return self.has_attribute(EnumStyleOptions)

  @property
  def contentPadding(self):
    """
    Description:
    -----------
    Add padding to the text content used when calculating the optimal width for a table.

    Related Pages:

			https://datatables.net/reference/option/columns.contentPadding
    """
    return self._attrs["contentPadding"]

  @contentPadding.setter
  def contentPadding(self, val):
    self._attrs["contentPadding"] = val

  @property
  def defaultContent(self):
    """
    Description:
    -----------
    Set default, static, content for a column.

    Related Pages:

			https://datatables.net/reference/option/columns.defaultContent
    """
    return self._attrs["defaultContent"]

  @defaultContent.setter
  def defaultContent(self, val):
    self._attrs["defaultContent"] = val

  @property
  def name(self):
    """
    Description:
    -----------
    When working with DataTables' API, it is very common to want to be able to address individual columns so you can work with them (you wish to sum the numeric content of a column for example). DataTables has two basic methods of addressing columns:

    Related Pages:

			https://datatables.net/reference/option/columns.name
    """
    return self._attrs["name"]

  @name.setter
  def name(self, val):
    self._attrs["name"] = val

  @property
  def title(self):
    """
    Description:
    -----------
    Set the column title.

    Related Pages:

			https://datatables.net/reference/option/columns.title
    """
    return self._attrs["title"]

  @title.setter
  def title(self, val):
    self._attrs["title"] = val

  @property
  def style(self):
    return self._attrs["style"]

  @style.setter
  def style(self, val):
    self._attrs["style"] = val

  @property
  def orderable(self):
    """
    Description:
    -----------
    Using this parameter, you can remove the end user's ability to order upon a column.
    This might be useful for generated content columns, for example if you have 'Edit' or 'Delete' buttons in the table.

    Related Pages:

			https://datatables.net/reference/option/columns.orderable
    """
    return self._attrs["orderable"]

  @orderable.setter
  def orderable(self, val):
    self._attrs["orderable"] = val

  @property
  def orderData(self):
    """
    Description:
    -----------
    Define multiple column ordering as the default order for a column.

    Related Pages:

			https://datatables.net/reference/option/columns.orderData
    """
    return self._attrs["orderData"]

  @orderData.setter
  def orderData(self, val):
    self._attrs["orderData"] = val

  @property
  def orderDataType(self):
    """
    Description:
    -----------
    Live DOM sorting type assignment.

    Related Pages:

			https://datatables.net/reference/option/columns.orderDataType
    """
    return self._attrs["orderDataType"]

  @orderDataType.setter
  def orderDataType(self, val):
    self._attrs["orderDataType"] = val

  @property
  def orderSequence(self):
    """
    Description:
    -----------
    You can control the default ordering direction, and even alter the behaviour of the order handler (i.e. only allow ascending sorting etc) using this parameter.

    Related Pages:

			https://datatables.net/reference/option/columns.orderSequence
    """
    return self._attrs["orderSequence"]

  @orderSequence.setter
  def orderSequence(self, val):
    self._attrs["orderSequence"] = val

  @property
  def render(self):
    """
    Description:
    -----------
    This property will modify the data that is used by DataTables for various operations as it is read from the data source. columns.render can be considered to be the the read only companion to columns.data which is read / write (and therefore more complex)

    Related Pages:

			https://datatables.net/reference/option/columns.render
    """
    return self._attrs["render"]

  @render.setter
  def render(self, val):
    self._attrs["render"] = val

  @property
  def data(self):
    return self._attrs["data"]

  @data.setter
  def data(self, val):
    self._attrs["data"] = val

  @property
  def searchable(self):
    """
    Description:
    -----------
    Using this parameter, you can define if DataTables should include this column in the filterable data in the table. You may want to use this option to disable search on generated columns such as 'Edit' and 'Delete' buttons for example.

    Related Pages:

			https://datatables.net/reference/option/columns.searchable
    """
    return self._attrs["searchable"]

  @searchable.setter
  def searchable(self, val):
    self._attrs["searchable"] = val

  @property
  def visible(self):
    """
    Description:
    -----------
    Enable or disable the display of this column.

    Related Pages:

			https://datatables.net/reference/option/columns.visible
    """
    return self._attrs["visible"]

  @visible.setter
  def visible(self, val):
    self._attrs["visible"] = val

  @property
  def width(self):
    """
    Description:
    -----------
    Column width assignment.

    Related Pages:

			https://datatables.net/reference/option/columns.width
    """
    return self._attrs["width"]

  @width.setter
  def width(self, val):
    self._attrs["width"] = val


class AOColumns(DataClass):

  def null(self):
    pass

  def add_order_sequence(self):
    pass


class Ajax(DataClass):

  @property
  def url(self):
    """
    https://datatables.net/manual/server-side

    :return:
    """
    return self._attrs["url"]

  @url.setter
  def url(self, val):
    self._attrs["url"] = val

  @property
  def type(self):
    """
    https://datatables.net/manual/server-side

    :return:
    """
    return self._attrs["type"]

  @type.setter
  def type(self, val):
    self._attrs["type"] = val


class Language(DataClass):

  @property
  def decimal(self):
    """
    Description:
    -----------
    Decimal place character.

    Related Pages:

			https://datatables.net/reference/option/language.decimal
    """
    return self._attrs["decimal"]

  @decimal.setter
  def decimal(self, val):
    self._attrs["decimal"] = val

  @property
  def url(self):
    """
    https://datatables.net/manual/i18n

    :return:
    """
    return self._attrs["url"]

  @url.setter
  def url(self, val):
    self._attrs["url"] = val

  @property
  def thousands(self):
    return self._attrs["thousands"]

  @thousands.setter
  def thousands(self, val):
    self._attrs["thousands"] = val


class Search(DataClass):

  @property
  def caseInsensitive(self):
    """
    Description:
    -----------
    Flag to indicate if the filtering should be case insensitive or not.

    Related Pages:

			https://datatables.net/reference/option/search.caseInsensitive
    """
    return self._attrs["caseInsensitive"]

  @caseInsensitive.setter
  def caseInsensitive(self, val):
    self._attrs["caseInsensitive"] = val

  @property
  def search(self):
    """
    Description:
    -----------
    The search option allows the way DataTables performs filtering to be set during the initialisation, and to set an initial global filter.

    Related Pages:

			https://datatables.net/reference/option/search
    """
    return self._attrs["search"]

  @search.setter
  def search(self, val):
    self._attrs["search"] = val

  @property
  def regex(self):
    """
    Description:
    -----------
    Regular expressions can be used to build fantastically complex filtering terms, but also it is perfectly valid for users to enter characters such as * into the filter, so a decision needs to be made if you wish to escape regular expression special characters or not.
    This option controls that ability in DataTables.

    Related Pages:

			https://datatables.net/reference/option/search.regex
    """
    return self._attrs["regex"]

  @regex.setter
  def regex(self, val):
    self._attrs["regex"] = val

  @property
  def smart(self):
    """
    Description:
    -----------
    DataTables' built-in filtering is "smart" in that it breaks the user's input into individual words and then matches those words in any position and in any order in the table (rather than simple doing a simple string compare).

    Related Pages:

			https://datatables.net/reference/option/search.smart
    """
    return self._attrs["smart"]

  @smart.setter
  def smart(self, val):
    self._attrs["smart"] = val


class TableConfig(DataClass):

  @property
  @packageImport('datatables-autoFill', 'datatables-autoFill')
  def autoFill(self):
    from epyk.core.html.tables.exts import DtAutoFill

    return self.sub_data("autoFill", DtAutoFill.AutoFill)

  @property
  def autoWidth(self):
    """
    Description:
    -----------
    Enable or disable automatic column width calculation.
    This can be disabled as an optimisation (it takes a finite amount of time to calculate the widths) if the tables widths are passed in using

    Related Pages:

			https://datatables.net/reference/option/autoWidth
    """
    return self._attrs["autoWidth"]

  @autoWidth.setter
  def autoWidth(self, val):
    self._attrs["autoWidth"] = val

  @property
  @packageImport('datatables-buttons', 'datatables-buttons')
  def buttons(self):
    """
    Description:
    -----------
    A common UI paradigm to use with interactive tables is to present buttons that will trigger some action - that may be to alter the table's state, modify the data in the table, gather the data from the table or even to activate some external process.
    Showing such buttons is an interface that end users are comfortable with, making them feel at home with the table.

    Related Pages:

			https://datatables.net/extensions/buttons/
    """
    from epyk.core.html.tables.exts import DtButtons
    self._attrs['dom'] = 'B<"clear">lfrtip'
    return self.sub_data("buttons", DtButtons.Buttons)

  @property
  @packageImport('datatables-buttons', 'datatables-buttons')
  def colReorder(self):
    """

    :return:
    """
    from epyk.core.html.tables.exts import DtColReorder
    return self.sub_data("colReorder", DtColReorder.ColReorder)

  @property
  def deferRender(self):
    """
    Description:
    -----------
    By default, when DataTables loads data from an Ajax or Javascript data source (ajax and data respectively)

    Related Pages:

			https://datatables.net/reference/option/deferRender

    """
    return self._attrs["deferRender"]

  @deferRender.setter
  def deferRender(self, val):
    self._attrs["deferRender"] = val


  @property
  @packageImport('datatables-fixed-header', 'datatables-fixed-header')
  def fixedHeader(self):
    """
    Description:
    -----------
    When displaying tables with a particularly large amount of data shown on each page, it can be useful to have the table's header and / or footer fixed to the top or bottom of the scrolling window.
    This lets your users quickly determine what each column refers to rather than needing to scroll back to the top of the table.

    Related Pages:

			https://datatables.net/extensions/fixedheader/
    """
    from epyk.core.html.tables.exts import DtFixedHeader
    return self.sub_data("fixedHeader", DtFixedHeader.FixedHeater)

  @property
  @packageImport('datatables-fixed-columns','datatables-fixed-columns')
  def fixedColumns(self):
    """

    :return:
    """
    from epyk.core.html.tables.exts import DtFixedColumns
    return self.sub_data("fixedColumns", DtFixedColumns.FixedColumns)

  @property
  @packageImport('datatables-keytable', 'datatables-keytable')
  def keys(self):
    """

    :rtype: DtFixedColumns.FixedColumns
    """
    from epyk.core.html.tables.exts import DtKeyTable
    return self.sub_data("keys", DtKeyTable.KeyTable)

  @property
  def lengthChange(self):
    """
    Description:
    -----------
    When pagination is enabled, this option will control the display of an option for the end user to change the number of records to be shown per page.
    The options shown in the list are controlled by the lengthMenu configuration option.

    Related Pages:

			https://datatables.net/reference/option/lengthChange
    """
    return self._attrs["lengthChange"]

  @lengthChange.setter
  def lengthChange(self, val):
    self._attrs["lengthChange"] = val

  @property
  def columnDefs(self):
    """
    Description:
    -----------
    Very similar to columns, this parameter allows you to assign specific options to columns in the table, although in this case the column options defined can be applied to one or more columns. Additionally, not every column need be specified, unlike columns.

    Related Pages:

			https://datatables.net/reference/option/columnDefs
    """
    return self.sub_data_enum("columnDefs", ColumnDef)

  @property
  def columns(self):
    """
    Description:
    -----------
    The columns option in the initialisation parameter allows you to define details about the way individual columns behave. For a full list of column options that can be set, please see the related parameters below.

    Related Pages:

			https://datatables.net/reference/option/columns
    """
    return self.sub_data_enum("columns", Column)

  @property
  def language(self):
    """
    Description:
    -----------
    Language configuration options for DataTables

    Related Pages:

			https://datatables.net/reference/option/
    """
    return self.sub_data("language", Language)

  @property
  def ajax(self):
    """
    Description:
    -----------
    DataTables can obtain the data that it is to display in the table body from a number of sources, including from an Ajax data source, using this initialisation parameter.
    As with other dynamic data sources, arrays or objects can be used for the data source for each row, with columns.data employed to read from specific object properties.

    Related Pages:

			https://datatables.net/reference/option/ajax
    """
    return self._attrs["ajax"]

  @ajax.setter
  def ajax(self, val):
    self._attrs["ajax"] = val

  @property
  def processing(self):
    """
    Description:
    -----------
    Enable or disable the display of a 'processing' indicator when the table is being processed (e.g. a sort).
    This is particularly useful for tables with large amounts of data where it can take a noticeable amount of time to sort the entries.

    Related Pages:

			https://datatables.net/reference/option/processing
    """
    return self._attrs["processing"]

  @processing.setter
  def processing(self, val):
    self._attrs["processing"] = val

  @property
  def search(self):
    """
    Description:
    -----------
    The search option allows the way DataTables performs filtering to be set during the initialisation, and to set an initial global filter.

    Related Pages:

			https://datatables.net/reference/option/search
    """
    return self.sub_data_enum("search", Search)

  @property
  def serverSide(self):
    """
    Description:
    -----------
    By default DataTables operates in client-side processing mode, but can be switched to server-side processing mode using this option.
    Server-side processing is useful when working with large data sets (typically >50'000 records) as it means a database engine can be used to perform the sorting etc calculations - operations that modern database engines are highly optimised for, allowing use of DataTables with massive data sets (millions of rows).

    Related Pages:

			https://datatables.net/reference/option/serverSide
    """
    return self._attrs["serverSide"]

  @serverSide.setter
  def serverSide(self, val):
    self._attrs["serverSide"] = val

  @property
  def deferLoading(self):
    """
    Description:
    -----------
    When using server-side processing, the default mode of operation for DataTables is to simply throw away any data that currently exists in the table and make a request to the server to get the first page of data to display.
    This is fine for an empty table, but if you already have the first page of data displayed in the plain HTML, it is a waste of resources.
    As such, this option exists to allow you to instruct DataTables to not make that initial request, rather it will use the data already on the page (no sorting etc will be applied to it).

    Related Pages:

			https://datatables.net/reference/option/deferLoading
    """
    return self._attrs["deferLoading"]

  @deferLoading.setter
  def deferLoading(self, val):
    self._attrs["deferLoading"] = val

  @property
  def destroy(self):
    """
    Description:
    -----------
    Initialise a new DataTable as usual, but if there is an existing DataTable which matches the selector, it will be destroyed and replaced with the new table.
    This can be useful if you want to change a property of the table which cannot be altered through the API.

    Related Pages:

			https://datatables.net/reference/option/destroy
    """
    return self._attrs["destroy"]

  @destroy.setter
  def destroy(self, val):
    self._attrs["destroy"] = val

  @property
  def displayStart(self):
    """
    Description:
    -----------
    Define the starting point for data display when using DataTables with pagination

    Related Pages:

			https://datatables.net/reference/option/displayStart
    """
    return self._attrs["displayStart"]

  @displayStart.setter
  def displayStart(self, val):
    self._attrs["displayStart"] = val

  @property
  def dom(self):
    """
    Description:
    -----------
    DataTables will add a number of elements around the table to both control the table and show additional information about it.
    The position of these elements on screen are controlled by a combination of their order in the document (DOM) and the CSS applied to the elements.
    This parameter is used to control their ordering and additional mark-up surrounding them in the DOM.

    Related Pages:

			https://datatables.net/reference/option/dom
    """
    return self._attrs["dom"]

  @dom.setter
  def dom(self, val):
    self._attrs["dom"] = val

  @property
  def data(self):
    """
    Description:
    -----------
    DataTables can obtain the data it is to display in the table's body from a number of sources, including being passed in as an array of row data using this initialisation parameter.
    As with other dynamic data sources, arrays or objects can be used for the data source for each row, with columns.data employed to read from specific object properties.

    Related Pages:

			https://datatables.net/reference/option/data
    """
    return self._attrs["data"]

  @data.setter
  def data(self, val):
    self._attrs["data"] = val

  @property
  def paging(self):
    """
    Description:
    -----------
    DataTables can split the rows in tables into individual pages, which is an efficient method of showing a large number of records in a small space.
    The end user is provided with controls to request the display of different data as the navigate through the data.
    This feature is enabled by default, but if you wish to disable it, you may do so with this parameter.

    Related Pages:

			https://datatables.net/reference/option/paging
    """
    return self._attrs["paging"]

  @paging.setter
  def paging(self, val):
    self._attrs["paging"] = val

  @property
  def info(self):
    """
    Description:
    -----------
    When this option is enabled, Datatables will show information about the table including information about filtered data if that action is being performed.
    This option allows that feature to be enabled or disabled.

    Related Pages:

			https://datatables.net/reference/option/info
    """
    return self._attrs["info"]

  @info.setter
  def info(self, val):
    self._attrs["info"] = val

  @property
  def orderCellsTop(self):
    """
    Description:
    -----------
    Allows control over whether DataTables should use the top (true) unique cell that is found for a single column, or the bottom (false - default) to attach the default order listener.
    This is useful when using complex headers.

    Related Pages:

			https://datatables.net/reference/option/orderCellsTop
    """
    return self._attrs["orderCellsTop"]

  @orderCellsTop.setter
  def orderCellsTop(self, val):
    self._attrs["orderCellsTop"] = val

  @property
  def orderClasses(self):
    """
    Description:
    -----------
    DataTables highlight the columns which are used to order the content in the table's body by adding a class to the cells in that column, which in turn has CSS applied to those classes to highlight those cells.

    Related Pages:

			https://datatables.net/reference/option/orderClasses
    """
    return self._attrs["orderClasses"]

  @orderClasses.setter
  def orderClasses(self, val):
    self._attrs["orderClasses"] = val

  @property
  def orderFixed(self):
    """
    Description:
    -----------
    The option works in tandem with the order option which provides an initial ordering state for the table which can then be modified by the user clicking on column headings, while the ordering specified by this option will always be applied to the table, regardless of user interaction.

    Related Pages:

			https://datatables.net/reference/option/orderFixed
    """
    return self._attrs["orderFixed"]

  @orderFixed.setter
  def orderFixed(self, val):
    self._attrs["orderFixed"] = val

  @property
  def orderMulti(self):
    """
    Description:
    -----------
    When ordering is enabled (ordering), by default DataTables allows users to sort multiple columns by shift clicking upon the header cell for each column.
    Although this can be quite useful for users, it can also increase the complexity of the order, potentiality increasing the processing time of ordering the data.
    Therefore, this option is provided to allow this shift-click multiple column ability.

    Related Pages:

			https://datatables.net/reference/option/orderMulti
    """
    return self._attrs["orderMulti"]

  @orderMulti.setter
  def orderMulti(self, val):
    self._attrs["orderMulti"] = val

  @property
  def ordering(self):
    """
    Description:
    -----------
    Enable or disable ordering of columns - it is as simple as that! DataTables, by default, allows end users to click on the header cell for each column, ordering the table by the data in that column.
    The ability to order data can be disabled using this option.

    Related Pages:

			https://datatables.net/reference/option/ordering
    """
    return self._attrs["ordering"]

  @ordering.setter
  def ordering(self, val):
    self._attrs["ordering"] = val

  def order(self, column, direction):
    """
    Description:
    -----------
    If ordering is enabled (ordering), then DataTables will perform a first pass order during initialisation.
    Using this parameter you can define which column(s) the order is performed upon, and the ordering direction.
    The order must be an array of arrays, each inner array comprised of two elements:

    Related Pages:

			https://datatables.net/reference/option/order

    Attributes:
    ----------
    :param column: String. The column name
    :param direction: String the direction (asc, desc)
    """
    if not 'order' in self._attrs:
      self._attrs["order"] = []
    self._attrs["order"].append([column, direction])
    return self

  @property
  def aoColumns(self):
    return self.sub_data_enum("aoColumns", AOColumns)

  @property
  @packageImport('datatables-responsive', 'datatables-responsive')
  def responsive(self):
    """
    Description:
    -----------
    In the modern world of responsive web design tables can often cause a particular problem for designers due to their row based layout.
    Responsive is an extension for DataTables that resolves that problem by optimising the table's layout for different screen sizes through the dynamic insertion and removal of columns from the table.

    Related Pages:

			https://datatables.net/extensions/responsive/

    :rtype: DtResponsive.Responsive
    """
    from epyk.core.html.tables.exts import DtResponsive
    return self.sub_data("responsive", DtResponsive.Responsive)

  @property
  def stateSave(self):
    """
    Description:
    -----------
    Enable or disable state saving.
    When enabled aDataTables will store state information such as pagination position, display length, filtering and sorting.
    When the end user reloads the page the table's state will be altered to match what they had previously set up.

    Related Pages:

			https://datatables.net/reference/option/stateSave
    """
    return self._attrs["stateSave"]

  @stateSave.setter
  def stateSave(self, val):
    self._attrs["stateSave"] = val

  @property
  def pagingType(self):
    return self._attrs["pagingType"]

  @pagingType.setter
  def pagingType(self, val):
    self._attrs["pagingType"] = val

  @property
  def scrollY(self):
    """
    Description:
    -----------
    Enable vertical scrolling.
    Vertical scrolling will constrain the DataTable to the given height, and enable scrolling for any data which overflows the current viewport. This can be used as an alternative to paging to display a lot of data in a small area (although paging and scrolling can both be enabled at the same time if desired).

    Related Pages:

			https://datatables.net/reference/option/scrollY
    """
    return self._attrs["scrollY"]

  @scrollY.setter
  def scrollY(self, val):
    self._attrs["scrollY"] = val

  @property
  def scrollX(self):
    """
    Description:
    -----------
    Enable horizontal scrolling.
    When a table is too wide to fit into a certain layout, or you have a large number of columns in the table, you can enable horizontal (x) scrolling to show the table in a viewport, which can be scrolled.

    Related Pages:

			https://datatables.net/reference/option/scrollX
    """
    return self._attrs["scrollX"]

  @scrollX.setter
  def scrollX(self, val):
    self._attrs["scrollX"] = val

  @property
  def scrollCollapse(self):
    """
    Description:
    -----------
    When vertical (y) scrolling is enabled through the use of the scrollY option, DataTables will force the height of the table's viewport to the given height at all times (useful for layout).
    However, this can look odd when filtering data down to a small data set, and the footer is left "floating" further down.

    Related Pages:

			https://datatables.net/reference/option/scrollCollapse
    """
    return self._attrs["scrollCollapse"]

  @scrollCollapse.setter
  def scrollCollapse(self, val):
    self._attrs["scrollCollapse"] = val

  @property
  def displayLength(self):
    return self._attrs["displayLength"]

  @displayLength.setter
  def displayLength(self, val):
    self._attrs["displayLength"] = val

  @property
  def scrollCollapse(self):
    return self._attrs["scrollCollapse"]

  @scrollCollapse.setter
  def scrollCollapse(self, val):
    self._attrs["scrollCollapse"] = val

  @property
  def lengthMenu(self):
    """
    Description:
    -----------
    This parameter allows you to readily specify the entries in the length drop down select list that DataTables shows when pagination is enabled. It can be either:

    Related Pages:

			https://datatables.net/reference/option/lengthMenu
    """
    return self._attrs["lengthMenu"]

  @lengthMenu.setter
  def lengthMenu(self, val):
    self._attrs["lengthMenu"] = val

  @property
  def select(self):
    """
    https://datatables.net/manual/options

    :return:
    """
    return self._attrs["scrollCollapse"]

  @select.setter
  def select(self, val):
    self._attrs["select"] = val

  @property
  def pageLength(self):
    """
    Description:
    -----------
    Number of rows to display on a single page when using pagination.

    Related Pages:

			https://datatables.net/reference/option/pageLength
    """
    return self._attrs["pageLength"]

  @pageLength.setter
  def pageLength(self, val):
    self._attrs["pageLength"] = val

  @property
  def pagingType(self):
    """
    Description:
    -----------
    The pagination option of DataTables will display a pagination control below the table (by default, its position can be changed using dom and CSS) with buttons that the end user can use to navigate the pages of the table.
    Which buttons are shown in the pagination control are defined by the option given here.

    Related Pages:

			https://datatables.net/reference/option/pagingType
    """
    return self._attrs["pagingType"]

  @pagingType.setter
  def pagingType(self, val):
    self._attrs["pagingType"] = val

  @property
  def renderer(self):
    """
    Description:
    -----------
    DataTables adds complex components to your HTML page, such as the pagination control.
    The business logic used to calculate what information should be displayed (what buttons in the case of the pagination buttons) is core to DataTables and generally doesn't vary how the buttons are actually displayed based on the styling requirements of the page.

    Related Pages:

			https://datatables.net/reference/option/renderer
    """
    return self._attrs["renderer"]

  @renderer.setter
  def renderer(self, val):
    self._attrs["renderer"] = val

  @property
  def retrieve(self):
    """
    Description:
    -----------
    Retrieve the DataTables object for the given selector.
    Note that if the table has already been initialised, this parameter will cause DataTables to simply return the object that has already been set up - it will not take account of any changes you might have made to the initialisation object passed to DataTables

    Related Pages:

			https://datatables.net/reference/option/retrieve
    """
    return self._attrs["retrieve"]

  @retrieve.setter
  def retrieve(self, val):
    self._attrs["retrieve"] = val

  @property
  def rowId(self):
    """
    Description:
    -----------
    It can often be useful to have a id attribute on each tr element in a DataTable for row selection and data source identification, particularly when using events.

    Related Pages:

			https://datatables.net/reference/option/rowId
    """
    return self._attrs["rowId"]

  @rowId.setter
  def rowId(self, val):
    self._attrs["rowId"] = val

  @property
  @packageImport('datatables-row-group', 'datatables-row-group')
  def rowGroup(self):
    """

    :return:
    """
    from epyk.core.html.tables.exts import DtFixedColumns
    return self.sub_data("rowGroup", DtFixedColumns.FixedColumns)

  @property
  def rowsGroup(self):
    """
    The Datatables feature plugin that groups rows (merge cells vertically) in according to specified columns.
    It's inspired by [fnFakeRowspan] (https://datatables.net/plug-ins/api/fnFakeRowspan) DataTables plugin.

    https://github.com/ashl1/datatables-rowsgroup
    """
    return self._attrs["rowsGroup"]

  @rowsGroup.setter
  def rowsGroup(self, val):
    """

    :param val:
    """
    self._report.jsImports.add('datatables-rows-group')
    self._attrs["rowsGroup"] = val

  @property
  @packageImport('datatables-select', 'datatables-select')
  def select(self):
    """
    Description:
    -----------
    Select adds item selection capabilities to a DataTable. Items can be rows, columns or cells, which can be selected independently, or together. Item selection can be particularly useful in interactive tables where users can perform some action on the table, such as editing rows or marking items to perform an action on.

    Related Pages:

			https://datatables.net/reference/option/select
    """
    from epyk.core.html.tables.exts import DtSelect
    return self.sub_data("select", DtSelect.Select)

  @property
  @packageImport('datatables-scroller', 'datatables-scroller')
  def scroller(self):
    """
    Description:
    -----------
    Scroller is a virtual rendering plug-in for DataTables which allows large datasets to be drawn on screen very quickly.
    What the virtual rendering means is that only the visible portion of the table (and a bit to either side to make the scrolling smooth) is drawn, while the scrolling container gives the visual impression that the whole table is visible.

    Related Pages:

			https://datatables.net/extensions/scroller/
    """
    from epyk.core.html.tables.exts import DtScroller
    return self.sub_data("scroller", DtScroller.Scroller)

  @property
  @packageImport('datatables-searchPanes', 'datatables-searchPanes')
  def searchPanes(self):
    """
    Description:
    -----------
    SearchPanes adds panes to the DataTable with the capability to search the DataTable by selecting rows in the panes.
    This is very useful when it comes to adding a more accessible searching feature and custom search capabilities.
    SearchPanes can search DataTables for multiple values that have been selected across multiple panes.
    They also provide the ability to define custom search functions which cannot be achieved through a simple searchBox.


    Related Pages:

			https://datatables.net/extensions/searchpanes
    """
    from epyk.core.html.tables.exts import DtSearchPanes
    self._attrs['dom'] = 'Prftip'
    return self.sub_data('searchPanes', DtSearchPanes.SearchPanes)

  @property
  def searchCols(self):
    """
    Description:
    -----------
    Basically the same as the search option, but in this case for individual columns, rather than the global filter, this option defined the filtering to apply to the table during initialisation.

    Related Pages:

			https://datatables.net/reference/option/searchCols
    """
    return self._attrs["searchCols"]

  @searchCols.setter
  def searchCols(self, val):
    self._attrs["searchCols"] = val

  @property
  def searchDelay(self):
    """
    Description:
    -----------
    The built-in DataTables global search (by default at the top right of every DataTable) will instantly search the table on every keypress when in client-side processing mode and reduce the search call frequency automatically to 400mS when in server-side processing mode.

    Related Pages:

			https://datatables.net/reference/option/searchDelay
    """
    return self._attrs["searchDelay"]

  @searchDelay.setter
  def searchDelay(self, val):
    self._attrs["searchDelay"] = val

  @property
  def stateDuration(self):
    """
    Description:
    -----------
    Duration for which the saved state information is considered valid. After this period has elapsed the state will be returned to the default.

    Related Pages:

			https://datatables.net/reference/option/stateDuration
    """
    return self._attrs["stateDuration"]

  @stateDuration.setter
  def stateDuration(self, val):
    self._attrs["stateDuration"] = val

  @property
  def stripeClasses(self):
    """
    Description:
    -----------
    An array of CSS classes that should be applied to displayed rows, in sequence. This array may be of any length, and DataTables will apply each class sequentially, looping when required.

    Related Pages:

			https://datatables.net/reference/option/stripeClasses
    """
    return self._attrs["stripeClasses"]

  @stripeClasses.setter
  def stripeClasses(self, val):
    self._attrs["stripeClasses"] = val

  @property
  def tabIndex(self):
    """
    Description:
    -----------
    By default DataTables allows keyboard navigation of the table (sorting, paging, and filtering) by adding a tabindex attribute to the required elements.

    Related Pages:

			https://datatables.net/reference/option/tabIndex
    """
    return self._attrs["tabIndex"]

  @tabIndex.setter
  def tabIndex(self, val):
    self._attrs["tabIndex"] = val



"""
CSS Style module for the Pivot components
"""

from epyk.core.css.styles.classes import CssStyle


class CssPivotHead(CssStyle.Style):
  _attrs = {'font-size': '8pt', 'padding': '5px'}
  classname = "pvtTable"
  _selectors = {'child': "tr th"}

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[-1], 'background-color': self.rptObj.theme.colors[0],
              #'border': '1px solid %s' % self.rptObj.theme.colors[3]
              })


class CssPivotCells(CssStyle.Style):
  _attrs = {'font-size': '8pt', 'padding': '5px'}
  classname = "pvtTable"
  _selectors = {'child': "tr td"}

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[-1], 'border': '1px solid %s' % self.rptObj.theme.colors[3]}, important=True)
    self.hover.css({'color': self.rptObj.theme.greys[-1], 'background-color': self.rptObj.theme.colors[2]}, important=True)


class CssPivotLabel(CssStyle.Style):
  classname = "pvtColLabel, .pvtAxisLabel, .pvtRowLabel, .pvtTotalLabel,  th"

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[-1], 'background-color': self.rptObj.theme.colors[1],
              'border': '1px solid %s' % self.rptObj.theme.colors[5]}, important=True)


class CssPivotAxis(CssStyle.Style):
  _attrs = {'background': 'red'}
  classname = "pvtAxisContainer"
  _selectors = {"child": "li span.pvtAttr"}

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[-1], #'background-color': self.rptObj.theme.colors[1],
              #'border': '1px solid %s' % self.rptObj.theme.colors[1]
              }, important=True)


class CssPivotFilterBox(CssStyle.Style):
  classname = "pvtAxisContainer"

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[1], #'background-color': self.rptObj.theme.colors[3],
              #'border': '1px solid %s' % self.rptObj.theme.colors[1]}
              })


class CssPivotFilterVals(CssStyle.Style):
  classname = "pvtVals"

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[-1],#'border': '1px solid %s' % self.rptObj.theme.colors[1]
              })


class CssPivotFilterBoxPopUp(CssStyle.Style):
  classname = "pvtFilterBox"

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[1],
              #'border': '1px solid %s' % self.rptObj.theme.colors[1]
              })



"""

"""

from epyk.core.js import Js
from epyk.core.js import JsUtils


class JsCookies(object):
  class __internal(object):
    _context = {}

  def __init__(self, src=None):
    self.src = src if src is not None else self.__internal()

  def set(self, jsKey, jsData, jsDataKey=None, isPyData=True, jsFnc=None):
    """

    Related Pages:

			https//www.w3schools.com/js/js_cookies.asp

    :return:
    """
    jsData = JsUtils.jsConvert(jsData, jsDataKey, isPyData, jsFnc)
    if self.src._context.get('cookies') is None:
      self.src._context['cookies'] = True
      return "document.cookies = {'%s': %s}" % (jsKey, jsData)

    return "document.cookies['%s'] = %s" % (jsKey, jsData)

  def get(self, jsData=None, jsDataKey=None, isPyData=True, jsFnc=None):
    """

    :return:
    """
    if jsData is None:
      return Js.JsJson().parse("decodeURIComponent(document.cookies)", isPyData=False)

    jsData = JsUtils.jsConvert(jsData, jsDataKey, isPyData, jsFnc)
    return Js.JsJson().parse("decodeURIComponent(document.cookies)['%s']" % jsData, isPyData=False)




from epyk.core.js.packages import packageImport
from epyk.core.html.options import Options

from epyk.core.js import JsUtils


class OptionsInput(Options):

  def css(self, attrs):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param attrs: Dictionary. CSS attributes
    """
    self._config(attrs)
    return self

  @property
  def disabled(self):
    """
    Description:
    ------------
    The rows attribute specifies the visible height of a text area, in lines.

    Related Pages:

			https://www.w3schools.com/tags/att_rows.asp
    """
    return self._report.attr.get('disabled', False)

  @disabled.setter
  def disabled(self, value):
    if not value and "disabled" in self._report.attr:
      del self._report.attr["disabled"]

    else:
      self._report.set_attrs({"disabled": value})

  @packageImport("accounting")
  def formatMoney(self, symbol="", digits=0, thousand_sep=".", decimal_sep=","):
    """
    Description:
    -----------
    Format any number into currency

    Related Pages:
http://openexchangerates.github.io/accounting.js/

    Attributes:
    ----------
    :param symbol: String. custom symbol
    :param digits: Integer. Number of digit
    :param thousand_sep: String. The thousand separator
    :param decimal_sep: String. The decimal separator
    """
    self._config({"symbol": symbol, 'digit': digits, "thousand": thousand_sep, "decimal": decimal_sep})
    return self

  @packageImport("accounting")
  def formatNumber(self, digit=0, thousand_sep="."):
    """
    Description:
    -----------
    Format a number with custom precision and localisation

    Related Pages:
http://openexchangerates.github.io/accounting.js/

    Attributes:
    ----------
    :param digit: Integer. Number of digit
    :param thousand_sep: String. The thousand separator
    """
    self._config({'digit': digit, "thousand": thousand_sep})
    return self

  @packageImport("accounting")
  def toFixed(self, digit=0):
    """
    Description:
    -----------
    Better rounding for floating point numbers

    Related Pages:
http://openexchangerates.github.io/accounting.js/

    Attributes:
    ----------
    :param digit: Integer. Number of digit
    """
    self._config(digit)
    return self

  @property
  def maxlength(self):
    """
    Description:
    ------------
    Specifies the maximum number of characters allowed in the text area

    Related Pages:

			https://www.w3schools.com/tags/att_textarea_maxlength.asp
    """
    return self._report.attr.get('maxlength')

  @maxlength.setter
  def maxlength(self, num):
    self._report.set_attrs({"maxlength": num})

  @property
  def name(self):
    """
    Description:
    ------------
    Specifies a name for a text area

    Related Pages:

			https://www.w3schools.com/tags/att_textarea_name.asp
    """
    return self._report.attr.get('name')

  @name.setter
  def name(self, num):
    self._report.set_attrs({"name": num})

  @property
  def placeholder(self):
    """
    Description:
    ------------
    The placeholder attribute specifies a short hint that describes the expected value of an input field (e.g. a sample value or a short description of the expected format).

    Related Pages:

			https://www.w3schools.com/tags/att_input_placeholder.asp
    """
    return self._report.attr.get('placeholder', "")

  @placeholder.setter
  def placeholder(self, value):
    self._report.set_attrs({"placeholder": value})

  @property
  def required(self):
    """
    Description:
    ------------
    Specifies that a text area is required/must be filled out

    Related Pages:

			https://www.w3schools.com/tags/att_textarea_required.asp
    """
    return self._report.attr.get('required')

  @required.setter
  def required(self, value):
    if not value and "required" in self._report.attr:
      del self._report.attr["required"]

    else:
      self._report.set_attrs({"required": value})

  @property
  def wrap(self):
    """
    Description:
    ------------
    Specifies how the text in a text area is to be wrapped when submitted in a form

    Related Pages:

			https://www.w3schools.com/tags/att_textarea_wrap.asp
    """
    return self._report.attr.get('wrap', "")

  @wrap.setter
  def wrap(self, value):
    self._report.set_attrs({"wrap": value})

  @property
  def spellcheck(self):
    """
    Description:
    ------------
    The spellcheck attribute specifies whether the element is to have its spelling and grammar checked or not.

    Related Pages:

			https://www.w3schools.com/tags/att_global_spellcheck.asp
    """
    return self._report.attr.get('spellcheck', False)

  @spellcheck.setter
  def spellcheck(self, value):
    self._report.set_attrs({"spellcheck": value})

  @property
  def readonly(self):
    """
    Description:
    ------------
    The readonly attribute is a boolean attribute.

    Related Pages:

			https://www.w3schools.com/tags/att_input_readonly.asp
    """
    return self._report.attr.get('readOnly', False)

  @readonly.setter
  def readonly(self, value):
    if not value and "readOnly" in self._report.attr:
      del self._report.attr["readOnly"]

    else:
      self._report.set_attrs({"readOnly": value})

  @property
  def reset(self):
    """
    Description:
    ------------
    """
    return self.get(False)

  @reset.setter
  def reset(self, bool):
    self.set(bool)

  @property
  def select(self):
    """
    Description:
    ------------
    """
    return self.get(False)

  @select.setter
  def select(self, bool):
    self.set(bool)


class OptionsInputRange(OptionsInput):

  @property
  def output(self):
    """
    Description:
    ------------
    """
    return self._attrs.get('output', True)

  @output.setter
  def output(self, bool):
    self.set(bool)


class OptionsInputInteger(OptionsInput):

  @property
  def quantity(self):
    """
    Description:
    ------------
    """
    return self.get(False)

  @quantity.setter
  def quantity(self, bool):
    self.set(bool)
    if bool:
      self._report.quantity()


class OptionsTimePicker(OptionsInput):

  @property
  def timeFormat(self):
    """
    Description:
    ------------
    string The format of the time string displayed in the input and the menu items in the combobox. Available modifiers are:

    Related Pages:

			https://timepicker.co/options/
    """
    return self._config_get(True)

  @timeFormat.setter
  def timeFormat(self, value):
    self._config(value)

  @property
  def minTime(self):
    """
    Description:
    ------------
    A Date object or string. Only the time parts (getHours, getMinutes) of the object are important.
    Time entries before minTime won't be displayed/allowed.

    Related Pages:

			https://timepicker.co/options/
    """
    return self._config_get(True)

  @minTime.setter
  def minTime(self, value):
    self._config(value)

  @property
  def minHour(self):
    """
    Description:
    ------------
    int. Time entries with an 24-hour part before minHour won't be displayed/allowed. Ignored if minTime is set.

    Related Pages:

			https://timepicker.co/options/
    """
    return self._config_get(True)

  @minHour.setter
  def minHour(self, number):
    self._config(number)

  @property
  def minMinutes(self):
    """
    Description:
    ------------
    int Time entries with minutes part before minMinutes won't be displayed/allowed. Ignored if minTime is set.

    Related Pages:

			https://timepicker.co/options/
    """
    return self._config_get(True)

  @minMinutes.setter
  def minMinutes(self, number):
    self._config(number)

  @property
  def maxTime(self):
    """
    Description:
    ------------
    A Date object or string. Only the time parts (getHours, getMinutes) of the object are important. Time entries after minTime won't be displayed/allowed.

    Related Pages:

			https://timepicker.co/options/
    """
    return self._config_get(True)

  @maxTime.setter
  def maxTime(self, number):
    self._config(number)

  @property
  def maxHour(self):
    """
    Description:
    ------------
    int. Time entries with an 24-hour part after maxHour won't be displayed/allowed. Ignored if maxTime is set.

    Related Pages:

			https://timepicker.co/options/
    """
    return self._config_get(True)

  @maxHour.setter
  def maxHour(self, number):
    self._config(number)

  @property
  def maxMinutes(self):
    """
    Description:
    ------------
    int. Time entries with minutes part after maxHour won't be displayed/allowed. Ignored if maxTime is set.

    Related Pages:

			https://timepicker.co/options/
    """
    return self._config_get(True)

  @maxMinutes.setter
  def maxMinutes(self, number):
    self._config(number)

  @property
  def startTime(self):
    """
    Description:
    ------------
    A Date object or string. The time of the first item in the combobox when the input field is empty.
    If the input field is not empty the first item will be the next allowed time entry.

    Related Pages:

			https://timepicker.co/options/
    """
    return self._config_get(True)

  @startTime.setter
  def startTime(self, value):
    self._config(value)

  @property
  def startHour(self):
    """
    Description:
    ------------
    int The 24-hour part of the first item in the combobox when the input field is emptye.
    If input field is not empty the first item will be the next allowed time entry. Ignored if startTime is set.

    Related Pages:

			https://timepicker.co/options/
    """
    return self._config_get(True)

  @startHour.setter
  def startHour(self, number):
    self._config(number)

  @property
  def startMinutes(self):
    """
    Description:
    ------------
    int The minutes part of the first item in the combobox when the input field is emptye.
    If input field is not empty the first item will be the next allowed time entry. Ignored if startTime is set.

    Related Pages:

			https://timepicker.co/options/
    """
    return self._config_get(True)

  @startMinutes.setter
  def startMinutes(self, number):
    self._config(number)

  @property
  def interval(self):
    """
    Description:
    ------------
    int Separation in minutes between time entries in the dropdown menu.

    Related Pages:

			https://timepicker.co/options/
    """
    return self._config_get(15)

  @interval.setter
  def interval(self, value):
    self._config(value)

  @property
  def dynamic(self):
    """
    Description:
    ------------
    boolean If a date is already selected and dynamic is true, the items in the dropdown will be arranged so that the first item is chronologically right after the selected time entry.

    Related Pages:

			https://timepicker.co/options/
    """
    return self._config_get(True)

  @dynamic.setter
  def dynamic(self, bool):
    self._config(bool)

  @property
  def scrollbar(self):
    """
    Description:
    ------------
    boolean Whether the scrollbars should be displayed or not.

    Related Pages:

			https://timepicker.co/options/
    """
    return self._config_get(True)

  @scrollbar.setter
  def scrollbar(self, bool):
    self._config(bool)

  @property
  def zindex(self):
    """
    Description:
    ------------
    int The value for the z-index property of the timepicker's container <div>.

    Related Pages:

			https://timepicker.co/options/
    """
    return self._config_get(True)

  @zindex.setter
  def zindex(self, number):
    self._config(number)


class OptionAutoComplete(Options):

  @property
  def appendTo(self):
    """
    Description:
    ------------
    Which element the menu should be appended to.
    When the value is null, the parents of the input field will be checked for a class of ui-front.
    If an element with the ui-front class is found, the menu will be appended to that element.
    Regardless of the value, if no element is found, the menu will be appended to the body.

    Related Pages:

			https://api.jqueryui.com/autocomplete/#option-appendTo
    """
    return self._config_get(None)

  @appendTo.setter
  def appendTo(self, value):
    self._config(value)

  @property
  def autoFocus(self):
    """
    Description:
    ------------
    If set to true the first item will automatically be focused when the menu is shown.

    Related Pages:

			https://api.jqueryui.com/autocomplete/#option-autoFocus
    """
    return self._config_get(False)

  @autoFocus.setter
  def autoFocus(self, value):
    self._config(value)

  @property
  def classes(self):
    """
    Description:
    ------------
    Specify additional classes to add to the widget's elements.
    Any of classes specified in the Theming section can be used as keys to override their value.
    To learn more about this option, check out the learn article about the classes option.

    Related Pages:

			https://api.jqueryui.com/autocomplete/#option-classes
    """
    return self._config_get([])

  @classes.setter
  def classes(self, value):
    self._config(value)

  @property
  def delay(self):
    """
    Description:
    ------------
    The delay in milliseconds between when a keystroke occurs and when a search is performed.
    A zero-delay makes sense for local data (more responsive), but can produce a lot of load for remote data, while being less responsive.

    Related Pages:

			https://api.jqueryui.com/autocomplete/#option-delay
    """
    return self._config_get(300)

  @delay.setter
  def delay(self, value):
    self._config(value)

  @property
  def disabled(self):
    """
    Description:
    ------------
    Disables the autocomplete if set to true.

    Related Pages:

			https://api.jqueryui.com/autocomplete/#option-delay
    """
    return self._config_get(True)

  @disabled.setter
  def disabled(self, value):
    self._config(value)

  @property
  def minLength(self):
    """
    Description:
    ------------
    The minimum number of characters a user must type before a search is performed.
    Zero is useful for local data with just a few items, but a higher value should be used when a single character search could match a few thousand items.

    Related Pages:

			https://api.jqueryui.com/autocomplete/#option-minLength
    """
    return self._config_get(0)

  @minLength.setter
  def minLength(self, value):
    self._config(value)

  def position(self, my="center", at="center", of="window"):
    """
    Description:
    ------------
    Specifies where the dialog should be displayed when opened. The dialog will handle collisions such that as much of the dialog is visible as possible.

    Related Pages:

			https://api.jqueryui.com/dialog/#option-position
    """
    self._config({"my": my, "at": at, "of": of})
    return self

  @property
  def source(self):
    """
    Description:
    ------------
    Defines the data to use, must be specified.

    Related Pages:

			https://api.jqueryui.com/autocomplete/#option-source
    """
    return self._config_get([])

  @source.setter
  def source(self, value):
    self._config(value)


class OptionsDatePicker(OptionsInput):

  @property
  def altField(self):
    """
    Description:
    ------------
    An input element that is to be updated with the selected date from the datepicker. Use the altFormat option to change the format of the date within this field. Leave as blank for no alternate field.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-altField
    """
    return self._config_get("")

  @altField.setter
  def altField(self, value):
    self._config(value)

  @property
  def altFormat(self):
    """
    Description:
    ------------
    The dateFormat to be used for the altField option. This allows one date format to be shown to the user for selection purposes, while a different format is actually sent behind the scenes. For a full list of the possible formats see the formatDate function

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-altFormat
    """
    return self._config_get("")

  @altFormat.setter
  def altFormat(self, value):
    self._config(value)

  @property
  def appendText(self):
    """
    Description:
    ------------
    The text to display after each date field, e.g., to show the required format.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-appendText
    """
    return self._config_get("")

  @appendText.setter
  def appendText(self, value):
    self._config(value)

  @property
  def autoSize(self):
    """
    Description:
    ------------
    Set to true to automatically resize the input field to accommodate dates in the current dateFormat.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-autoSize
    """
    return self._config_get(False)

  @autoSize.setter
  def autoSize(self, value):
    self._config(value)

  @property
  def beforeShow(self):
    """
    Description:
    ------------
    A function that takes an input field and current datepicker instance and returns an options object to update the datepicker with. It is called just before the datepicker is displayed.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-beforeShow
    """
    return self._config_get(None)

  @beforeShow.setter
  def beforeShow(self, value):
    self._config(value)

  @property
  def beforeShowDay(self):
    """
    Description:
    ------------
    A function that takes a date as a parameter and must return an array with:

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-beforeShowDay
    """
    return self._config_get(None)

  @beforeShowDay.setter
  def beforeShowDay(self, value):
    self._config(value)

  @property
  def buttonImage(self):
    """
    Description:
    ------------
    A URL of an image to use to display the datepicker when the showOn option is set to "button" or "both". If set, the buttonText option becomes the alt value and is not directly displayed.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-buttonImage
    """
    return self._config_get("")

  @buttonImage.setter
  def buttonImage(self, value):
    self._config(value)

  @property
  def buttonImageOnly(self):
    """
    Description:
    ------------
    Whether the button image should be rendered by itself instead of inside a button element. This option is only relevant if the buttonImage option has also been set.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-buttonImageOnly
    """
    return self._config_get(None)

  @buttonImageOnly.setter
  def buttonImageOnly(self, value):
    self._config(value)

  @property
  def buttonText(self):
    """
    Description:
    ------------
    The text to display on the trigger button. Use in conjunction with the showOn option set to "button" or "both".

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-buttonText
    """
    return self._config_get("...")

  @buttonText.setter
  def buttonText(self, value):
    self._config(value)

  @property
  def calculateWeek(self):
    """
    Description:
    ------------
    A function to calculate the week of the year for a given date. The default implementation uses the ISO 8601 definition: weeks start on a Monday; the first week of the year contains the first Thursday of the year.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-calculateWeek
    """
    return self._config_get(None)

  @calculateWeek.setter
  def calculateWeek(self, value):
    self._config(value)

  @property
  def changeMonth(self):
    """
    Description:
    ------------
    Whether the month should be rendered as a dropdown instead of text.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-changeMonth
    """
    return self._config_get(False)

  @changeMonth.setter
  def changeMonth(self, value):
    self._config(value)

  @property
  def changeYear(self):
    """
    Description:
    ------------
    Whether the year should be rendered as a dropdown instead of text. Use the yearRange option to control which years are made available for selection.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-changeYear
    """
    return self._config_get(False)

  @changeYear.setter
  def changeYear(self, value):
    self._config(value)

  @property
  def closeText(self):
    """
    Description:
    ------------
    The text to display for the close link. Use the showButtonPanel option to display this button.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-closeText
    """
    return self._config_get("Done")

  @closeText.setter
  def closeText(self, value):
    self._config(value)

  @property
  def constrainInput(self):
    """
    Description:
    ------------
    When true, entry in the input field is constrained to those characters allowed by the current dateFormat option.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-constrainInput
    """
    return self._config_get(True)

  @constrainInput.setter
  def constrainInput(self, value):
    self._config(value)

  @property
  def currentText(self):
    """
    Description:
    ------------
    The text to display for the current day link. Use the showButtonPanel option to display this button.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-currentText
    """
    return self._config_get("Today")

  @currentText.setter
  def currentText(self, value):
    self._config(value)

  @property
  def dateFormat(self):
    """
    Description:
    ------------
    The format for parsed and displayed dates. For a full list of the possible formats see the formatDate function.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-dateFormat
    """
    return self._config_get("mm/dd/yy")

  @dateFormat.setter
  def dateFormat(self, value):
    self._config(value)

  @property
  def dayNames(self):
    """
    Description:
    ------------
    The list of long day names, starting from Sunday, for use as requested via the dateFormat option.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-dayNames
    """
    return self._config_get([])

  @dayNames.setter
  def dayNames(self, value):
    self._config(value)

  @property
  def dayNamesMin(self):
    """
    Description:
    ------------
    The list of minimised day names, starting from Sunday, for use as column headers within the datepicker.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-dayNamesMin
    """
    return self._config_get([])

  @dayNamesMin.setter
  def dayNamesMin(self, value):
    self._config(value)

  @property
  def dayNamesShort(self):
    """
    Description:
    ------------
    The list of abbreviated day names, starting from Sunday, for use as requested via the dateFormat option.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-dayNamesShort
    """
    return self._config_get([])

  @dayNamesShort.setter
  def dayNamesShort(self, value):
    self._config(value)

  @property
  def defaultDate(self):
    """
    Description:
    ------------
    Set the date to highlight on first opening if the field is blank. Specify either an actual date via a Date object or as a string in the current dateFormat, or a number of days from today (e.g. +7) or a string of values and periods ('y' for years, 'm' for months, 'w' for weeks, 'd' for days, e.g. '+1m +7d'), or null for today.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-defaultDate
    """
    return self._config_get(None)

  @defaultDate.setter
  def defaultDate(self, value):
    self._config(value)

  @property
  def duration(self):
    """
    Description:
    ------------
    Control the speed at which the datepicker appears, it may be a time in milliseconds or a string representing one of the three predefined speeds ("slow", "normal", "fast").

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-duration
    """
    return self._config_get("normal")

  @duration.setter
  def duration(self, value):
    self._config(value)

  @property
  def inline(self):
    """
    Description:
    ------------
    """
    return self.get(False)

  @inline.setter
  def inline(self, bool):
    self.set(bool)

  @property
  def firstDay(self):
    """
    Description:
    ------------
    Set the first day of the week: Sunday is 0, Monday is 1, etc.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-firstDay
    """
    return self._config_get(0)

  @firstDay.setter
  def firstDay(self, value):
    self._config(value)

  @property
  def gotoCurrent(self):
    """
    Description:
    ------------
    When true, the current day link moves to the currently selected date instead of today.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-gotoCurrent
    """
    return self._config_get(False)

  @gotoCurrent.setter
  def gotoCurrent(self, value):
    self._config(value)

  @property
  def hideIfNoPrevNext(self):
    """
    Description:
    ------------
    Normally the previous and next links are disabled when not applicable (see the minDate and maxDate options). You can hide them altogether by setting this attribute to true.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-hideIfNoPrevNext
    """
    return self._config_get(False)

  @hideIfNoPrevNext.setter
  def hideIfNoPrevNext(self, value):
    self._config(value)

  @property
  def isRTL(self):
    """
    Description:
    ------------
    Whether the current language is drawn from right to left.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-isRTL
    """
    return self._config_get(False)

  @isRTL.setter
  def isRTL(self, value):
    self._config(value)

  @property
  def maxDate(self):
    """
    Description:
    ------------
    The maximum selectable date. When set to null, there is no maximum.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-maxDate
    """
    return self._config_get(None)

  @maxDate.setter
  def maxDate(self, value):
    self._config(value)

  @property
  def minDate(self):
    """
    Description:
    ------------
    The minimum selectable date. When set to null, there is no minimum.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-minDate
    """
    return self._config_get(None)

  @minDate.setter
  def minDate(self, value):
    self._config(value)

  @property
  def monthNames(self):
    """
    Description:
    ------------
    The list of full month names, for use as requested via the dateFormat option.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-monthNames
    """
    return self._config_get([])

  @monthNames.setter
  def monthNames(self, value):
    self._config(value)

  @property
  def monthNamesShort(self):
    """
    Description:
    ------------
    The list of abbreviated month names, as used in the month header on each datepicker and as requested via the dateFormat option.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-monthNamesShort
    """
    return self._config_get([])

  @monthNamesShort.setter
  def monthNamesShort(self, value):
    self._config(value)

  @property
  def navigationAsDateFormat(self):
    """
    Description:
    ------------
    Whether the currentText, prevText and nextText options should be parsed as dates by the formatDate function, allowing them to display the target month names for example.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-navigationAsDateFormat
    """
    return self._config_get(None)

  @navigationAsDateFormat.setter
  def navigationAsDateFormat(self, value):
    self._config(value)

  @property
  def nextText(self):
    """
    Description:
    ------------
    The text to display for the next month link. With the standard ThemeRoller styling, this value is replaced by an icon.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-nextText
    """
    return self._config_get("Next")

  @nextText.setter
  def nextText(self, value):
    self._config(value)

  @property
  def numberOfMonths(self):
    """
    Description:
    ------------
    The number of months to show at once.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-numberOfMonths
    """
    return self._config_get(1)

  @numberOfMonths.setter
  def numberOfMonths(self, value):
    self._config(value)

  @property
  def onChangeMonthYear(self):
    """
    Description:
    ------------
    Called when the datepicker moves to a new month and/or year. The function receives the selected year, month (1-12), and the datepicker instance as parameters. this refers to the associated input field.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-onChangeMonthYear
    """
    return self._config_get(None)

  @onChangeMonthYear.setter
  def onChangeMonthYear(self, value):
    self._config(value)

  @property
  def onClose(self):
    """
    Description:
    ------------
    Called when the datepicker is closed, whether or not a date is selected. The function receives the selected date as text ("" if none) and the datepicker instance as parameters. this refers to the associated input field.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-onClose
    """
    return self._config_get(None)

  @onClose.setter
  def onClose(self, value):
    self._config(value)

  @property
  def onSelect(self):
    """
    Description:
    ------------
    Called when the datepicker is selected. The function receives the selected date as text and the datepicker instance as parameters. this refers to the associated input field.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-onSelect
    """
    return self._config_get(None)

  @onSelect.setter
  def onSelect(self, values):
    values = JsUtils.jsConvertFncs(values, toStr=True)
    self._config("function(dateText, inst){let data = dateText; %s}" % values)

  @property
  def prevText(self):
    """
    Description:
    ------------
    The text to display for the previous month link. With the standard ThemeRoller styling, this value is replaced by an icon.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-prevText
    """
    return self._config_get("Prev")

  @prevText.setter
  def prevText(self, value):
    self._config(value)

  @property
  def selectOtherMonths(self):
    """
    Description:
    ------------
    Whether days in other months shown before or after the current month are selectable. This only applies if the showOtherMonths option is set to true.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-selectOtherMonths
    """
    return self._config_get(False)

  @selectOtherMonths.setter
  def selectOtherMonths(self, value):
    self._config(value)

  @property
  def shortYearCutoff(self):
    """
    Description:
    ------------
    The cutoff year for determining the century for a date (used in conjunction with dateFormat 'y'). Any dates entered with a year value less than or equal to the cutoff year are considered to be in the current century, while those greater than it are deemed to be in the previous century.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-shortYearCutoff
    """
    return self._config_get("+10")

  @shortYearCutoff.setter
  def shortYearCutoff(self, value):
    self._config(value)

  @property
  def showAnim(self):
    """
    Description:
    ------------
    The name of the animation used to show and hide the datepicker. Use "show" (the default), "slideDown", "fadeIn", any of the jQuery UI effects. Set to an empty string to disable animation.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-showAnim
    """
    return self._config_get("show")

  @showAnim.setter
  def showAnim(self, value):
    self._config(value)

  @property
  def showButtonPanel(self):
    """
    Description:
    ------------
    Whether to display a button pane underneath the calendar. The button pane contains two buttons, a Today button that links to the current day, and a Done button that closes the datepicker. The buttons' text can be customized using the currentText and closeText options respectively.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-showButtonPanel
    """
    return self._config_get(False)

  @showButtonPanel.setter
  def showButtonPanel(self, value):
    self._config(value)

  @property
  def showCurrentAtPos(self):
    """
    Description:
    ------------
    When displaying multiple months via the numberOfMonths option, the showCurrentAtPos option defines which position to display the current month in.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-showCurrentAtPos
    """
    return self._config_get(0)

  @showCurrentAtPos.setter
  def showCurrentAtPos(self, value):
    self._config(value)

  @property
  def showMonthAfterYear(self):
    """
    Description:
    ------------
    Whether to show the month after the year in the header.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-showMonthAfterYear
    """
    return self._config_get(False)

  @showMonthAfterYear.setter
  def showMonthAfterYear(self, value):
    self._config(value)

  @property
  def showOn(self):
    """
    Description:
    ------------
    When the datepicker should appear. The datepicker can appear when the field receives focus ("focus"), when a button is clicked ("button"), or when either event occurs ("both").

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-showOn
    """
    return self._config_get("focus")

  @showOn.setter
  def showOn(self, value):
    self._config(value)

  @property
  def showOptions(self):
    """
    Description:
    ------------
    If using one of the jQuery UI effects for the showAnim option, you can provide additional properties for that animation using this option.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-showOptions
    """
    return self._config_get({})

  @showOptions.setter
  def showOptions(self, value):
    self._config(value)

  @property
  def showOtherMonths(self):
    """
    Description:
    ------------
    Whether to display dates in other months (non-selectable) at the start or end of the current month. To make these days selectable use the selectOtherMonths option.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-showOtherMonths
    """
    return self._config_get(False)

  @showOtherMonths.setter
  def showOtherMonths(self, value):
    self._config(value)

  @property
  def showWeek(self):
    """
    Description:
    ------------
    When true, a column is added to show the week of the year. The calculateWeek option determines how the week of the year is calculated. You may also want to change the firstDay option.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-showWeek
    """
    return self._config_get(False)

  @showWeek.setter
  def showWeek(self, value):
    self._config(value)

  @property
  def stepMonths(self):
    """
    Description:
    ------------
    Set how many months to move when clicking the previous/next links.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-stepMonths
    """
    return self._config_get(1)

  @stepMonths.setter
  def stepMonths(self, value):
    self._config(value)

  @property
  def weekHeader(self):
    """
    Description:
    ------------
    The text to display for the week of the year column heading. Use the showWeek option to display this column.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-weekHeader
    """
    return self._config_get("wk")

  @weekHeader.setter
  def weekHeader(self, value):
    self._config(value)

  @property
  def yearRange(self):
    """
    Description:
    ------------
    The range of years displayed in the year drop-down: either relative to today's year ("-nn:+nn"), relative to the currently selected year ("c-nn:c+nn"), absolute ("nnnn:nnnn"), or combinations of these formats ("nnnn:-nn"). Note that this option only affects what appears in the drop-down, to restrict which dates may be selected use the minDate and/or maxDate options.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-yearRange
    """
    return self._config_get("c-10:c+10")

  @yearRange.setter
  def yearRange(self, value):
    self._config(value)

  @property
  def yearSuffix(self):
    """
    Description:
    ------------
    Additional text to display after the year in the month headers.

    Related Pages:

			https://api.jqueryui.com/datepicker/#option-yearSuffix
    """
    return self._config_get("")

  @yearSuffix.setter
  def yearSuffix(self, value):
    self._config(value)


class OptionsTextarea(OptionsInput):

  @property
  def rows(self):
    """
    Description:
    ------------
    The rows attribute specifies the visible height of a text area, in lines.

    Related Pages:

			https://www.w3schools.com/tags/att_rows.asp
    """
    return self._report.attr.get('rows', "")

  @rows.setter
  def rows(self, value):
    self._report.set_attrs({"rows": value})




"""
CSS Style module for the Radio components
"""

from epyk.core.css.styles.classes import CssStyle

from epyk.core.css import Defaults as Defaults_css


class CssRadioButton(CssStyle.Style):
  _attrs = {'padding': '2px 5px', 'cursor': 'pointer', 'vertical-align': 'middle'}

  def customize(self):
    self.css({'border': '1px solid %s' % self.rptObj.theme.colors[3], 'font-size': '%spx' % Defaults_css.Font.size})


class CssRadioButtonSelected(CssStyle.Style):
  _attrs = {'padding': '2px 5px', 'cursor': 'pointer', 'vertical-align': 'middle'}

  def customize(self):
    self.css({'border': '1px solid %s' % self.rptObj.theme.success[1], 'color': self.rptObj.theme.success[1],
              'font-size': '%spx' % Defaults_css.Font.size})


class CssRadioSwitch(CssStyle.Style):
  _attrs = {'height': 0, 'width': 0, 'visibility': 'hidden'}


class CssRadioSwitchLabel(CssStyle.Style):
  _attrs = {'cursor': 'pointer', 'margin': '2px', 'text-indent': '-9999px', 'display': 'block', 'border-radius': '50px',
            'position': 'relative', 'top': '5px'}
  _after = {'content': "''", 'position': 'absolute', 'left': '5px', 'width': '15px', 'height': '15px',
            'border-radius': '20px', 'transition': '0.3s', 'margin': 'auto', 'top': '-2.5px'}

  def customize(self):
    self.css({'background': self.rptObj.theme.greys[3]})
    self.after.css({'background-color': self.rptObj.theme.success[1]})


class CssRadioSwitchChecked(CssStyle.Style):
  _after = {'left': 'calc(100% - 5px)', 'transform': 'translateX(-100%)'}
  _selectors = {'child': 'input:checked + label'}



'''
Module used as a wrapper to the Javascript C3 libraries

Related Pages:

		https://c3js.org/gettingstarted.html
  http://c3js.org/
This module is defined by a main class ** Chart **.

The constructor ::__init__
::onDocumentLoadVar
::onDocumentReady


Python / Javascript Events
::click
::mouseover
::mouseout


Pure Javascript Wrapper
Those function will be only used in **Javascript called** and they will return a piece of string which will be added in the
report to get the data later on in the Javascript layer. Python is just used here to put all the pieces together

The method to destroy the C3 chart ::jsDestroy
The method to group the different charts ::jsGroups

'''


from epyk.core.data.DataClass import DataClass

from epyk.core.html import Html

from epyk.core.js.packages import JsBillboard
from epyk.core.js.primitives import JsObjects
from epyk.core.js import JsUtils

from epyk.core.js.packages import JsD3


class Chart(Html.Html):
  name = 'Billboard Chart'
  requirements = ('billboard', )

  def __init__(self, report, width, height, htmlCode, options, profile):
    self.height = height[0]
    super(Chart, self).__init__(report, [], htmlCode=htmlCode, css_attrs={"width": width, "height": height},
                                profile=profile)
    self._d3, self._datasets, self._options, self._data_attrs, self._attrs = None, [], None, {}, {}
    self._options_init = options
    self.style.css.margin_top = 10

  @property
  def js(self):
    """
    Description:
    -----------
    JC3 reference API

    https://c3js.org/reference.html#api-show

    :return: A Javascript object

    :rtype: JsC3.C3
    """
    if self._js is None:
      self._js = JsBillboard.Billboard(self, varName=self.chartId, report=self._report)
    return self._js

  @property
  def chartId(self):
    """
    Description:
    -----------
    Return the Javascript variable of the chart
    """
    return "%s_obj" % self.htmlCode

  def click(self, jsFncs, profile=False):
    self.data.onclick(jsFncs, profile)
    return self

  @property
  def d3(self):
    if self._d3 is None:
      self._d3 = JsD3.D3Select(self._report, selector="d3.select('#%s')" % self.htmlCode, setVar=False)
    return self._d3

  def convert(self, data, options, profile=False):
    mod_name = __name__.split(".")[-1]
    constructors = self._report._props.setdefault("js", {}).setdefault("constructors", {})
    constructors[self.builder_name] = "function %s%sConvert(data, options){%s; return result}" % (
      mod_name, self.builder_name, self._js__convertor__)
    if isinstance(data, dict):
      # check if there is no nested HTML components in the data
      tmp_data = ["%s: %s" % (JsUtils.jsConvertData(k, None), JsUtils.jsConvertData(v, None)) for k, v in data.items()]
      js_data = "{%s}" % ",".join(tmp_data)
    else:
      js_data = JsUtils.jsConvertData(data, None)
    dfl_options, js_options = dict(self._options_init), []
    if options is not None:
      dfl_options.update(options)
    for k, v in dfl_options.items():
      if isinstance(v, dict):
        row = ["'%s': %s" % (s_k, JsUtils.jsConvertData(s_v, None)) for s_k, s_v in v.items()]
        js_options.append("'%s': {%s}" % (k, ", ".join(row)))
      else:
        if str(v).strip().startswith("function"):
          js_options.append("%s: %s" % (k, v))
        else:
          js_options.append("%s: %s" % (k, JsUtils.jsConvertData(v, None)))
    return "%s%sConvert(%s, %s)" % (mod_name, self.builder_name, js_data, "{%s}" % ",".join(js_options))

  def build(self, data=None, options=None, profile=False):
    if data:
      return '%(chartId)s.unload(); %(chartId)s.load(%(data)s)' % {'chartId': self.chartId, 'data': self.convert(data, options, profile)}

    return '%s = bb.generate(%s)' % (self.chartId, self.getCtx())

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<div %s></div>' % self.get_attrs(pyClassNames=self.style.get_classes())


class JsTick(DataClass):

  @property
  def format(self):
    return self._attrs["format"]

  @format.setter
  def format(self, val):
    self._attrs["format"] = val


class BBAxis(DataClass):

  @property
  def type(self):
    return self._attrs["type"]

  @type.setter
  def type(self, val):
    self._attrs["type"] = val

  @property
  def show(self):
    return self._attrs["show"]

  @show.setter
  def show(self, val):
    self._attrs["show"] = val

  @property
  def tick(self):
    return self.sub_data("x", JsTick)


class BBSelection(DataClass):
  @property
  def enabled(self):
    """
    Set data selection enabled.

    If this option is set true, we can select the data points and get/set its state of selection by API (e.g. select, unselect, selected).

    https://c3js.org/reference.html#data-selection-enabled
    """
    return self._attrs["enabled"]

  @enabled.setter
  def enabled(self, val):
    self._attrs["enabled"] = val

  @property
  def grouped(self):
    """
    """
    return self._attrs["grouped"]

  @grouped.setter
  def grouped(self, val):
    self._attrs["grouped"] = val

  @property
  def multiple(self):
    """
    """
    return self._attrs["multiple"]

  @multiple.setter
  def multiple(self, val):
    self._attrs["multiple"] = val

  @property
  def draggable(self):
    """
    """
    return self._attrs["draggable"]

  @draggable.setter
  def draggable(self, val):
    self._attrs["draggable"] = val

  @property
  def isselectable(self):
    """
    """
    return self._attrs["isselectable"]

  @isselectable.setter
  def isselectable(self, jsFnc):
    self._attrs["isselectable"] = jsFnc


class JsData(DataClass):
  @property
  def x(self):
    return self._attrs["x"]

  @x.setter
  def x(self, val):
    self._attrs["x"] = val

  @property
  def xs(self): return self._attrs["xs"]

  @xs.setter
  def xs(self, val): self._attrs["xs"] = val

  @property
  def xFormat(self): return self._attrs["xFormat"]

  @xFormat.setter
  def xFormat(self, val): self._attrs["xFormat"] = val

  @property
  def names(self): return self._attrs["names"]

  @names.setter
  def names(self, val): self._attrs["names"] = val

  @property
  def groups(self): return self._attrs["groups"]

  @groups.setter
  def groups(self, val): self._attrs["groups"] = val

  @property
  def axes(self): return self._attrs["axes"]

  @axes.setter
  def axes(self, val): self._attrs["axes"] = val

  @property
  def type(self): return self._attrs["type"]

  @type.setter
  def type(self, val): self._attrs["type"] = val

  def add_type(self, alias, type):
    if "types" not in self._attrs:
      self.types = {}
    self.types[alias] = type
    return self

  @property
  def types(self):
    """
    This setting overwrites data.type setting:
    line, spline, step, area...

    https://c3js.org/reference.html#data-types
    """
    return self._attrs["types"]

  @types.setter
  def types(self, val): self._attrs["types"] = val

  @property
  def labels(self): return self._attrs["labels"]

  @labels.setter
  def labels(self, val): self._attrs["labels"] = val

  @property
  def order(self): return self._attrs["order"]

  @order.setter
  def order(self, val): self._attrs["order"] = val

  @property
  def colors(self):
    if "colors" not in self._attrs:
      self._attrs["colors"] = {}
    return self._attrs["colors"]

  @colors.setter
  def colors(self, val): self._attrs["colors"] = val

  @property
  def columns(self):
    if 'columns' not in self._attrs:
      self._attrs["columns"] = []
    return self._attrs["columns"]

  @columns.setter
  def columns(self, val): self._attrs["columns"] = val

  @property
  def hide(self): return self._attrs["hide"]

  @hide.setter
  def hide(self, val): self._attrs["hide"] = val

  @property
  def selection(self):
    return self.sub_data("selection", BBSelection)

  def onclick(self, jsFncs, profile=False):
    self._attrs["onclick"] = JsObjects.JsObject.JsObject("function () { %s }" % JsUtils.jsConvertFncs(jsFncs, toStr=True))


class JsScales(DataClass):

  @property
  def rotated(self):
    return self._attrs["rotated"]

  @rotated.setter
  def rotated(self, val):
    self._attrs["rotated"] = val

  @property
  def x(self):
    return self.sub_data("x", BBAxis)

  @property
  def y(self):
    return self.sub_data("y", BBAxis)

  @property
  def y2(self):
    """

    https://c3js.org/reference.html#axis-y2-show

    :rtype: C3Axis
    """
    return self.sub_data("y2", BBAxis)


class BBGridLine(DataClass):
  @property
  def value(self):
    return self._attrs["value"]

  @value.setter
  def value(self, val):
    self._attrs["value"] = val

  @property
  def text(self):
    return self._attrs["text"]

  @text.setter
  def text(self, val):
    self._attrs["text"] = val

  @property
  def css_class(self):
    return self._attrs["class"]

  @css_class.setter
  def css_class(self, val):
    self._attrs["class"] = val

  @property
  def position(self):
    return self._attrs["position"]

  @position.setter
  def position(self, val):
    self._attrs["position"] = val


class C3GridAxis(DataClass):

  @property
  def show(self):
    return self._attrs["show"]

  @show.setter
  def show(self, val):
    self._attrs["show"] = val

  @property
  def lines(self):
    return self.sub_data_enum("lines", BBGridLine)


class BBGrid(DataClass):

  @property
  def x(self):
    return self.sub_data("x", C3GridAxis)

  @property
  def y(self):
    return self.sub_data("y", C3GridAxis)


class JsLegend(DataClass):
  pass


class JsTooltip(DataClass):
  pass


class JsSubchart(DataClass):
  pass


class JsZoom(DataClass):
  @property
  def enabled(self):
    return self._attrs["enabled"]

  @enabled.setter
  def enabled(self, val):
    self._attrs["enabled"] = val

  @property
  def type(self):
    return self._attrs["type"]

  @type.setter
  def type(self, val):
    self._attrs["type"] = val

  @property
  def rescale(self):
    return self._attrs["rescale"]

  @rescale.setter
  def rescale(self, val):
    self._attrs["rescale"] = val

  @property
  def extent(self):
    return self._attrs["extent"]

  @extent.setter
  def extent(self, val):
    self._attrs["extent"] = val


class BBPoints(DataClass):

  @property
  def show(self):
    return self._attrs["show"]

  @show.setter
  def show(self, val):
    self._attrs["show"] = val

  @property
  def r(self):
    return self._attrs["r"]

  @r.setter
  def r(self, val):
    self._attrs["r"] = val

  @property
  def focus(self):
    return self._attrs["focus"]

  @focus.setter
  def focus(self, val):
    self._attrs["focus"] = {"expand": val, 'enabled': True}

  @property
  def select(self):
    return self._attrs["select"]

  @select.setter
  def select(self, val):
    self._attrs["select"] = val


class ChartLine(Chart):
  _type = 'line'

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(ChartLine, self).__init__(report, width, height, htmlCode, options, profile)
    self._attrs["bindto"] = "#%s" % self.htmlCode

  def labels(self, labels, series_id='x'):
    self.data.x = series_id
    self.data.columns.append([series_id] + labels)

  def add_dataset(self, name, data, type=None):
    self.data.columns.append([name] + data)
    self.data.colors[name] = self._report.theme.colors[len(self.data.colors)]
    if type is None:
      self.data.add_type(name, self._type)
    return self._attrs

  @property
  def _js__convertor__(self):
    return '''
      if(data.python){ 
        result = {'columns': [], type: options.type};
        result['columns'].push(['x'].concat(data.labels));
        data.series.forEach(function(name, i){
          result['columns'].push([name].concat(data.datasets[i]));
        }); 
      } else {
        var temp = {}; var labels = []; var uniqLabels = {};
        options.y_columns.forEach(function(series){temp[series] = {}});
        data.forEach(function(rec){ 
          options.y_columns.forEach(function(name){
            if(rec[name] !== undefined){
              if (!(rec[options.x_column] in uniqLabels)){labels.push(rec[options.x_column]); uniqLabels[rec[options.x_column]] = true};
              temp[name][rec[options.x_column]] = rec[name]}})});
        columns = []; columns.push(['x'].concat(labels));
        options.y_columns.forEach(function(series){
          dataSet = [series];
          labels.forEach(function(x){
            if(temp[series][x] == undefined){dataSet.push(null)} else {dataSet.push(temp[series][x])}}); columns.push(dataSet)});
        var result = {columns: columns}
      }'''

  @property
  def axis(self):
    """

    :rtype: JsScales
    :return:
    """
    if not'axis' in self._attrs:
      self._attrs['axis'] = JsScales(self._report)
    return self._attrs['axis']

  @property
  def point(self):
    """

    :rtype: C3Points
    """
    if not 'point' in self._attrs:
      self._attrs['point'] = BBPoints(self._report)
    return self._attrs['point']

  @property
  def zoom(self):
    """

    :rtype: JsZoom
    """
    if not 'zoom' in self._attrs:
      self._attrs['zoom'] = JsZoom(self._report)
    return self._attrs['zoom']

  @property
  def data(self):
    """

    :rtype: JsScales
    :return:
    """
    if not 'data' in self._attrs:
      self._attrs['data'] = JsData(self._report)
    return self._attrs['data']

  @property
  def grid(self):
    """

    :rtype: JsScales
    :return:
    """
    if not 'grid' in self._attrs:
      self._attrs['grid'] = BBGrid(self._report)
    return self._attrs['grid']

  def getCtx(self):
    str_ctx = "{%s}" % ", ".join(["%s: %s" % (k, JsUtils.jsConvertData(v, None)) for k, v in self._attrs.items()])
    return str_ctx


class ChartSpline(ChartLine):
  _type = 'spline'


class ChartArea(ChartLine):
  _type = 'area'


class ChartBar(ChartLine):
  _type = 'bar'


class ChartScatter(ChartLine):
  _type = 'scatter'

  @property
  def _js__convertor__(self):
    return '''
      if(data.python){ 
        result = {'columns': [], type: options.type};
        result['columns'].push(['x'].concat(data.labels));
        data.series.forEach(function(name, i){
          result['columns'].push([name].concat(data.datasets[i]));
        });
      } else {
        var tempVal = {}; var tempX = {}; var labels = []; 
        options.y_columns.forEach(function(series){tempVal[series] = []; tempX[series +"_x"] = []});
        data.forEach(function(rec){ 
          options.y_columns.forEach(function(name){
            if(rec[name] !== undefined){
              if(!(rec[options.x_column] in tempVal[name])){tempVal[name] = [name, rec[name]]; tempX[name +"_x"] = [name +"_x", rec[options.x_column]]}
              else {tempVal[name].push(rec[name]); tempX[name +"_x"].push(rec[options.x_column])}}})});
        result = {'columns': [], 'xs': {}};
        options.y_columns.forEach(function(series){
          result.columns.push(tempVal[series]); result.columns.push(tempX[series +"_x"]); result.xs[series] = series +"_x"})
      }'''


class ChartPie(ChartLine):
  _type = 'pie'

  @property
  def _js__convertor__(self):
    return '''
      if(data.python){ 
        result = {'columns': [], type: options.type};
        result['columns'].push(['x'].concat(data.labels));
        data.series.forEach(function(name, i){
          result['columns'].push([name].concat(data.datasets[i]));
        }); 
      } else {
        var temp = {}; var labels = {};
        data.forEach(function(rec){ 
          if(!(rec[options.x_column] in temp)){temp[rec[options.x_column]] = {}};
          options.y_columns.forEach(function(name){
            labels[name] = true; 
            if(rec[name] !== undefined){
              if(!(name in temp[rec[options.x_column]])){temp[rec[options.x_column]][name] = rec[name]} else{temp[rec[options.x_column]][name] += rec[name]}}})});
        columns = []; var labels = Object.keys(labels); var count = 0;
        for(var series in temp){
          var values = [count]; count += 1;
          labels.forEach(function(label){
            if(temp[series][label] !== undefined){values.push(temp[series][label])} else{values.push(null)}});
          columns.push(values)};
        var result = {columns: columns};
      }'''

  def labels(self, labels, series_id='x'):
    self._labels = labels

  def add_dataset(self, name, values, type=None):
    for i, value in enumerate(values):
      self.data.columns.append([self._labels[i], value])
      self.data.colors[self._labels[i]] = self._report.theme.colors[i]
      if type is None:
        self.data.add_type(self._labels[i], self._type)
    return self._attrs


class ChartDonut(ChartPie):
  _type = 'donut'


class ChartGauge(ChartPie):
  _type = 'gauge'

  def build(self, data=None, options=None, profile=False):
    if data:
      return '%(chartId)s.load({columns: [["data", %(value)s]]})' % {'chartId': self.chartId, 'value': data}

    return '%s = bb.generate(%s)' % (self.chartId, self.getCtx())

  def add_dataset(self, name, value, type=None):
    self.data.columns.append(["data", value])
    self.data.colors["data"] = self._report.theme.colors[len(self.data.colors)]
    if type is None:
      self.data.add_type("data", self._type)
    return self._attrs


class ChartBubble(ChartLine):
  _type = 'bubble'


class ChartRadar(ChartLine):
  _type = 'radar'






from epyk.core.html import graph


class ChartGoogle(object):

  def __init__(self, context):
    self.parent = context
    self.chartFamily = "Google"

  def line(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    A line chart that is rendered within the browser using SVG or VML. Displays tooltips when hovering over points.

    Related Pages:

			https://developers.google.com/chart/interactive/docs/gallery/linechart

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    options = options or {'type': 'LineChart'}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'colors': self.parent.context.rptObj.theme.charts, 'attrs': {'fill': None}})
    data = self.parent.context.rptObj.data.google.y(record, y_columns, x_axis)
    line_chart = graph.GraphGoogle.ChartLine(self.parent.context.rptObj, data, width, height, htmlCode, options, profile)
    return line_chart

  def column(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    A column chart is a vertical bar chart rendered in the browser using SVG or VML, whichever is appropriate for the user's browser.
    Like all Google charts, column charts display tooltips when the user hovers over the data. For a horizontal version of this chart, see the bar chart.

    Related Pages:

      https://developers.google.com/chart/interactive/docs/gallery/columnchart

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    options = options or {'type': 'ColumnChart'}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'colors': self.parent.context.rptObj.theme.charts,
                    'attrs': {'fill': None}})
    data = self.parent.context.rptObj.data.google.y(record, y_columns, x_axis)
    line_chart = graph.GraphGoogle.ChartLine(self.parent.context.rptObj, data, width, height, htmlCode, options,
                                             profile)
    return line_chart

  def bar(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Google bar charts are rendered in the browser using SVG or VML, whichever is appropriate for the user's browser.
    Like all Google charts, bar charts display tooltips when the user hovers over the data. For a vertical version of this chart, see the column chart.

    Related Pages:

			https://developers.google.com/chart/interactive/docs/gallery/barchart

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    options = options or {'type': 'BarChart'}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'colors': self.parent.context.rptObj.theme.charts, 'attrs': {'fill': None}})
    data = self.parent.context.rptObj.data.google.y(record, y_columns, x_axis)
    line_chart = graph.GraphGoogle.ChartLine(self.parent.context.rptObj, data, width, height, htmlCode, options, profile)
    return line_chart

  def scatter(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Scatter charts plot points on a graph. When the user hovers over the points, tooltips are displayed with more information.

    Google scatter charts are rendered within the browser using SVG or VML depending on browser capabilities.

    Related Pages:

			https://developers.google.com/chart/interactive/docs/gallery/scatterchart

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    options = options or {'type': 'ScatterChart'}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'colors': self.parent.context.rptObj.theme.charts, 'attrs': {'fill': None}})
    data = self.parent.context.rptObj.data.google.y(record, y_columns, x_axis)
    line_chart = graph.GraphGoogle.ChartLine(self.parent.context.rptObj, data, width, height, htmlCode, options, profile)
    return line_chart

  def histogram(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    A histogram is a chart that groups numeric data into bins, displaying the bins as segmented columns.
    They're used to depict the distribution of a dataset: how often values fall into ranges.

    Google Charts automatically chooses the number of bins for you.
    All bins are equal width and have a height proportional to the number of data points in the bin.
    In other respects, histograms are similar to column charts.

    Related Pages:

			https://developers.google.com/chart/interactive/docs/gallery/histogram

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    options = options or {'type': 'Histogram'}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'colors': self.parent.context.rptObj.theme.charts, 'attrs': {'fill': None}})
    data = self.parent.context.rptObj.data.google.y(record, y_columns, x_axis)
    line_chart = graph.GraphGoogle.ChartLine(self.parent.context.rptObj, data, width, height, htmlCode, options, profile)
    return line_chart

  def area(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    An area chart that is rendered within the browser using SVG or VML. Displays tips when hovering over points.

    Related Pages:

			https://developers.google.com/chart/interactive/docs/gallery/areachart

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    options = options or {'type': 'AreaChart'}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'colors': self.parent.context.rptObj.theme.charts, 'attrs': {'fill': None}})
    data = self.parent.context.rptObj.data.google.y(record, y_columns, x_axis)
    line_chart = graph.GraphGoogle.ChartLine(self.parent.context.rptObj, data, width, height, htmlCode, options, profile)
    return line_chart

  def bubble(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    A bubble chart that is rendered within the browser using SVG or VML. Displays tips when hovering over bubbles.

    A bubble chart is used to visualize a data set with two to four dimensions.
    The first two dimensions are visualized as coordinates, the third as color and the fourth as size.

    Related Pages:

			https://developers.google.com/chart/interactive/docs/gallery/bubblechart

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    options = options or {'type': 'BubbleChart'}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'colors': self.parent.context.rptObj.theme.charts, 'attrs': {'fill': None}})
    data = self.parent.context.rptObj.data.google.y(record, y_columns, x_axis)
    line_chart = graph.GraphGoogle.ChartLine(self.parent.context.rptObj, data, width, height, htmlCode, options, profile)
    return line_chart

  def pie(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    A pie chart that is rendered within the browser using SVG or VML. Displays tooltips when hovering over slices.

    Related Pages:

			https://developers.google.com/chart/interactive/docs/gallery/piechart

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    options = options or {'type': 'PieChart'}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'colors': self.parent.context.rptObj.theme.charts, 'attrs': {'fill': None}})
    data = self.parent.context.rptObj.data.google.y(record, y_columns, x_axis)
    line_chart = graph.GraphGoogle.ChartLine(self.parent.context.rptObj, data, width, height, htmlCode, options, profile)
    return line_chart

  def treemap(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------

    Related Pages:

			https://developers.google.com/chart/interactive/docs/gallery/treemap

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    options = options or {'type': 'TreeMap'}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'colors': self.parent.context.rptObj.theme.charts, 'attrs': {'fill': None}})
    data = self.parent.context.rptObj.data.google.y(record, y_columns, x_axis)
    line_chart = graph.GraphGoogle.ChartLine(self.parent.context.rptObj, data, width, height, htmlCode, options, profile)
    return line_chart

  def candlestick(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    A candlestick chart is used to show an opening and closing value overlaid on top of a total variance. Candlestick charts are often used to show stock value behavior.
    In this chart, items where the opening value is less than the closing value (a gain) are drawn as filled boxes, and items where the opening value is more than the closing value (a loss) are drawn as hollow boxes.

    Related Pages:

			https://developers.google.com/chart/interactive/docs/gallery/candlestickchart

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    options = options or {'type': 'CandlestickChart'}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'colors': self.parent.context.rptObj.theme.charts, 'attrs': {'fill': None}})
    data = self.parent.context.rptObj.data.google.y(record, y_columns, x_axis)
    line_chart = graph.GraphGoogle.ChartLine(self.parent.context.rptObj, data, width, height, htmlCode, options, profile)

    return line_chart

  def gauge(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    A gauge with a dial, rendered within the browser using SVG or VML.

    Related Pages:

			https://developers.google.com/chart/interactive/docs/gallery/gauge

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    options = options or {'type': 'Gauge'}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'colors': self.parent.context.rptObj.theme.charts, 'attrs': {'fill': None}})
    data = self.parent.context.rptObj.data.google.y(record, y_columns, x_axis)
    line_chart = graph.GraphGoogle.ChartLine(self.parent.context.rptObj, data, width, height, htmlCode, options, profile)
    return line_chart

  def geo(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    A geochart is a map of a country, a continent, or a region with areas identified in one of three ways:

      The region mode colors whole regions, such as countries, provinces, or states.
      The markers mode uses circles to designate regions that are scaled according to a value that you specify.
      The text mode labels the regions with identifiers (e.g., "Russia" or "Asia").

    Related Pages:

			https://developers.google.com/chart/interactive/docs/gallery/geochart

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    """
    options = options or {'type': 'GeoChart'}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'colors': self.parent.context.rptObj.theme.charts, 'attrs': {'fill': None}})
    data = self.parent.context.rptObj.data.google.y(record, y_columns, x_axis)
    line_chart = graph.GraphGoogle.ChartLine(self.parent.context.rptObj, data, width, height, htmlCode, options, profile)
    return line_chart




from epyk.core.js.Js import JsConsole
from epyk.core.js.Js import JsWindow
from epyk.core.js.Js import JsBase
from epyk.core.js.Js import JsMaths

from epyk.core.css import Selector

from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsObjects

from epyk.core.js.objects.JsNodeDom import JsDoms
from epyk.core.js.objects.JsNodeDom import JsDomsList
from epyk.core.js.objects.JsEvents import Event


def selector(component=None):
  """
  Description:
  -----------
  CSS Selectors

  Related Pages:

			https://www.w3schools.com/cssref/css_selectors.asp

  Attributes:
  ----------
  :param component:
  """
  return Selector.Selector(component)


initEvent = Event().initEvent
createEvent = Event().createEvent
getEvent = Event().getEvent

console = JsConsole()

window = JsWindow.JsWindow()

alert = JsWindow.JsWindow().alert


def querySelectorAll(selector):
  """
  Description:
  ------------
  The querySelectorAll() method returns all elements in the document that matches a specified CSS selector(s), as a static NodeList object.

  Related Pages:

			https://www.w3schools.com/jsref/met_document_queryselectorall.asp

  Attributes:
  ----------
  :param selector: String. CSS selectors
  """
  return JsDomsList("document.querySelectorAll(%s)" % JsUtils.jsConvertData(selector, None), isPyData=False)


def querySelector(selector):
  """
  Description:
  ------------
  The querySelector() method returns the first element that matches a specified CSS selector(s) in the document.

  Related Pages:

			https://www.w3schools.com/jsref/met_document_queryselector.asp

  Attributes:
  ----------
  :param selector: String. CSS selectors
  """
  return JsDoms.get("document.querySelector(%s)" % JsUtils.jsConvertData(selector, None))


getElementById = JsBase.getElementById
getElementsByName = JsBase.getElementsByName
getElementsByTagName = JsBase.getElementsByTagName
getElementsByClassName = JsBase.getElementsByClassName

parseFloat = JsBase.parseFloat
parseInt = JsBase.parseInt
parseDate = JsBase.parseDate
typeof = JsBase.typeof

maths = JsMaths.JsMaths()


def comment(value):
  """
  Description:
  ------------
  Javascript Comment section.

  Related Pages:

			https://www.w3schools.com/js/js_comments.asp

  Attributes:
  ----------
  :param value: String. the Value
  """
  return JsObjects.JsVoid("/*%s*/" % value)


def var(name, value=None):
  """
  Description:
  ------------

  Hoisting is JavaScript's default behavior of moving declarations to the top.

  Attributes:
  ----------
  :param name:
  :param value:
  """
  if value is None:
    return JsObjects.JsObject.JsObject.get(name)

  return JsObjects.JsVoid("var %s = %s" % (name, JsUtils.jsConvertData(value, None)))


def let(name, value):
  """
  Description:
  ------------
  Redeclaring a variable using the let keyword can solve this problem.

  Redeclaring a variable inside a block will not redeclare the variable outside the block:

  Redeclaring a var variable with let, in the same scope, or in the same block, is not allowed:

  https://www.w3schools.com/js/js_let.asp

  Attributes:
  ----------
  :param name:
  :param value:
  """
  return JsObjects.JsVoid("let %s = %s" % (name, JsUtils.jsConvertData(value, None)))


def const(name, value=None):
  """
  Description:
  ------------
  The keyword const is a little misleading.

  It does NOT define a constant value. It defines a constant reference to a value.

  Redeclaring or reassigning an existing const variable, in the same scope, or in the same block, is not allowed:

  https://www.w3schools.com/js/js_const.asp

  Attributes:
  ----------
  :param name:
  :param value:
  """
  return JsObjects.JsVoid("const %s = %s" % (name, JsUtils.jsConvertData(value, None)))


debugger = JsObjects.JsVoid("debugger")



"""
This module will propose an implementation of a GPRC call.

THis module has been designed based on the examples available here: https://github.com/grpc/grpc/tree/v1.23.0
"""

import sys
import importlib


class DataGrpc(object):
  def __init__(self, service_name, path, module, host, port):
    """
    The path will be added to the python path

    :param service_name: The Service name (the class name in the python module)
    :param path: The path with the GRPC features
    :param module: The python module name for the service
    :param host: The service host name (e.g localhost)
    :param port: The service port
    """
    if not path in sys.path:
      sys.path.append(path)
    self.host, self.port, self.module = host, port, module.replace(".py", "")
    self.service_name, self._libs = service_name, {}
    if not module.endswith("_grpc"):
      raise Exception("Service module should ends with _grpc")

  @property
  def py(self):
    """
    Return the main module in the GRPC service for the data.
    Namely is return the module without the _grpc

    Example
    grpc.py.HelloRequest(name="Test 3")

    :return: A python module
    """
    return self.imp(self.module.replace("_grpc", ""))

  def imp(self, module):
    """
    Return any module from the GRPC service based on its module name

    Example
    grpc.imp(self.module)

    :param module: The python module name

    :return: A python module
    """
    module = module.replace(".py", "")
    if not module in self._libs:
      self._libs[module] = importlib.import_module(module)
    return self._libs[module]

  def request(self, method, data=None, options=None):
    """
    Run the GRPC call from the service definition in the init

    Example
    grpc.request("SayHello", data)

    :param method: The function name to be called for the request
    :param data: The data to be passed during the call
    :param options: The GRPC service call options

    :return:
    """
    import grpc

    with grpc.insecure_channel(target="%s:%s" % (self.host, self.port), options=options) as channel:
      stub = getattr(self.imp(self.module), self.service_name)(channel)
      if data is None:
        results = getattr(stub, method)()
      else:
        results = getattr(stub, method)(data)
      return results
